from datetime import datetime
from typing import Any, Dict, List, Optional

import numpy
import pandas as pd
from sqlalchemy import JSON, BigInteger, Column, ForeignKey, Index, Integer, Text
from sqlalchemy.orm import Session, relationship

from hummingbot.core.event.events import PositionAction
from hummingbot.model import HummingbotBase
from hummingbot.model.decimal_type_decorator import SqliteDecimal


class TradeFill(HummingbotBase):
    __tablename__ = "TradeFill"
    __table_args__ = (Index("tf_config_timestamp_index",
                            "config_file_path", "timestamp"),
                      Index("tf_market_trading_pair_timestamp_index",
                            "market", "symbol", "timestamp"),
                      Index("tf_market_base_asset_timestamp_index",
                            "market", "base_asset", "timestamp"),
                      Index("tf_market_quote_asset_timestamp_index",
                            "market", "quote_asset", "timestamp")
                      )

    config_file_path = Column(Text, nullable=False)
    strategy = Column(Text, nullable=False)
    market = Column(Text, primary_key=True, nullable=False)
    symbol = Column(Text, nullable=False)
    base_asset = Column(Text, nullable=False)
    quote_asset = Column(Text, nullable=False)
    timestamp = Column(BigInteger, nullable=False)
    order_id = Column(Text, ForeignKey("Order.id"), primary_key=True, nullable=False)
    trade_type = Column(Text, nullable=False)
    order_type = Column(Text, nullable=False)
    price = Column(SqliteDecimal(6), nullable=False)
    amount = Column(SqliteDecimal(6), nullable=False)
    leverage = Column(Integer, nullable=False, default=1)
    trade_fee = Column(JSON, nullable=False)
    trade_fee_in_quote = Column(SqliteDecimal(6))
    exchange_trade_id = Column(Text, primary_key=True, nullable=False)
    position = Column(Text, nullable=True, default=PositionAction.NIL.value)
    order = relationship("Order", back_populates="trade_fills")

    def __repr__(self) -> str:
        return f"TradeFill(config_file_path='{self.config_file_path}', strategy='{self.strategy}', " \
               f"market='{self.market}', symbol='{self.symbol}', base_asset='{self.base_asset}', " \
               f"quote_asset='{self.quote_asset}', timestamp={self.timestamp}, order_id='{self.order_id}', " \
               f"trade_type='{self.trade_type}', order_type='{self.order_type}', price={self.price}, " \
               f"amount={self.amount}, leverage={self.leverage}, trade_fee={self.trade_fee}, " \
               f"exchange_trade_id={self.exchange_trade_id}, position={self.position})"

    @staticmethod
    def get_trades(sql_session: Session,
                   strategy: str = None,
                   market: str = None,
                   trading_pair: str = None,
                   base_asset: str = None,
                   quote_asset: str = None,
                   trade_type: str = None,
                   order_type: str = None,
                   start_time: int = None,
                   end_time: int = None,
                   ) -> Optional[List["TradeFill"]]:
        filters = []
        if strategy is not None:
            filters.append(TradeFill.strategy == strategy)
        if market is not None:
            filters.append(TradeFill.market == market)
        if trading_pair is not None:
            filters.append(TradeFill.symbol == trading_pair)
        if base_asset is not None:
            filters.append(TradeFill.base_asset == base_asset)
        if quote_asset is not None:
            filters.append(TradeFill.quote_asset == quote_asset)
        if trade_type is not None:
            filters.append(TradeFill.trade_type == trade_type)
        if order_type is not None:
            filters.append(TradeFill.order_type == order_type)
        if start_time is not None:
            filters.append(TradeFill.timestamp >= start_time)
        if end_time is not None:
            filters.append(TradeFill.timestamp <= end_time)

        trades: Optional[List[TradeFill]] = (sql_session
                                             .query(TradeFill)
                                             .filter(*filters)
                                             .order_by(TradeFill.timestamp.asc())
                                             .all())
        return trades

    @classmethod
    def to_pandas(cls, trades: List):
        columns: List[str] = ["Id",
                              "Timestamp",
                              "Exchange",
                              "Market",
                              "Order_type",
                              "Side",
                              "Price",
                              "Amount",
                              "Leverage",
                              "Position",
                              "Age"]
        data = []
        for trade in trades:

            if trade.order is None:  # order creation update has not arrived yet
                age = pd.Timestamp(0, unit='s').strftime('%H:%M:%S')
            else:
                age = pd.Timestamp(int(trade.timestamp / 1e3 - trade.order.creation_timestamp / 1e3),
                                   unit='s').strftime('%H:%M:%S')
            data.append([
                trade.exchange_trade_id,
                datetime.fromtimestamp(int(trade.timestamp / 1e3)).strftime("%Y-%m-%d %H:%M:%S"),
                trade.market,
                trade.symbol,
                trade.order_type.lower(),
                trade.trade_type.lower(),
                trade.price,
                trade.amount,
                trade.leverage,
                trade.position,
                age,
            ])
        df = pd.DataFrame(data=data, columns=columns)
        df.set_index('Id', inplace=True)

        return df

    @staticmethod
    def to_bounty_api_json(trade_fill: "TradeFill") -> Dict[str, Any]:
        return {
            "market": trade_fill.market,
            "trade_id": trade_fill.exchange_trade_id,
            "price": numpy.format_float_positional(trade_fill.price),
            "quantity": numpy.format_float_positional(trade_fill.amount),
            "symbol": trade_fill.symbol,
            "trade_timestamp": trade_fill.timestamp,
            "trade_type": trade_fill.trade_type,
            "base_asset": trade_fill.base_asset,
            "quote_asset": trade_fill.quote_asset,
            "raw_json": {
                "trade_fee": trade_fill.trade_fee,
            }
        }

    @staticmethod
    def attribute_names_for_file_export():

        return [
            "exchange_trade_id",  # Keep the key attribute first in the list
            "config_file_path",
            "strategy",
            "market",
            "symbol",
            "base_asset",
            "quote_asset",
            "timestamp",
            "order_id",
            "trade_type",
            "order_type",
            "price",
            "amount",
            "leverage",
            "trade_fee",
            "trade_fee_in_quote",
            "position", ]



import asyncio
import importlib
import inspect
import os
from decimal import Decimal
from typing import Callable, Dict, List, Optional, Set

import pandas as pd
import yaml
from pydantic import Field, field_validator

from hummingbot.client import settings
from hummingbot.client.config.config_data_types import BaseClientModel
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.connector.markets_recorder import MarketsRecorder
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.common import MarketDict, PositionMode
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.data_feed.market_data_provider import MarketDataProvider
from hummingbot.exceptions import InvalidController
from hummingbot.remote_iface.mqtt import ETopicPublisher
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase
from hummingbot.strategy_v2.controllers.controller_base import ControllerBase, ControllerConfigBase
from hummingbot.strategy_v2.controllers.directional_trading_controller_base import (
    DirectionalTradingControllerConfigBase,
)
from hummingbot.strategy_v2.controllers.market_making_controller_base import MarketMakingControllerConfigBase
from hummingbot.strategy_v2.executors.data_types import PositionSummary
from hummingbot.strategy_v2.executors.executor_orchestrator import ExecutorOrchestrator
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executor_actions import (
    CreateExecutorAction,
    ExecutorAction,
    StopExecutorAction,
    StoreExecutorAction,
)
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo


class StrategyV2ConfigBase(BaseClientModel):
    """
    Base class for version 2 strategy configurations.
    """
    markets: MarketDict = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter markets in format 'exchange1.tp1,tp2:exchange2.tp1,tp2':",
            "prompt_on_new": True}
    )
    candles_config: List[CandlesConfig] = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter candle configs in format 'exchange1.tp1.interval1.max_records:exchange2.tp2.interval2.max_records':",
            "prompt_on_new": True,
        }
    )
    controllers_config: List[str] = Field(
        default=[],
        json_schema_extra={
            "prompt": "Enter controller configurations (comma-separated file paths), leave it empty if none: ",
            "prompt_on_new": True,
        }
    )

    @field_validator("controllers_config", mode="before")
    @classmethod
    def parse_controllers_config(cls, v):
        # Parse string input into a list of file pathsq
        if isinstance(v, str):
            if v == "":
                return []
            return [item.strip() for item in v.split(',') if item.strip()]
        if v is None:
            return []
        return v

    def load_controller_configs(self):
        loaded_configs = []
        for config_path in self.controllers_config:
            full_path = os.path.join(settings.CONTROLLERS_CONF_DIR_PATH, config_path)
            with open(full_path, 'r') as file:
                config_data = yaml.safe_load(file)

            controller_type = config_data.get('controller_type')
            controller_name = config_data.get('controller_name')

            if not controller_type or not controller_name:
                raise ValueError(f"Missing controller_type or controller_name in {config_path}")

            module_path = f"{settings.CONTROLLERS_MODULE}.{controller_type}.{controller_name}"
            module = importlib.import_module(module_path)

            config_class = next((member for member_name, member in inspect.getmembers(module)
                                 if inspect.isclass(member) and member not in [ControllerConfigBase,
                                                                               MarketMakingControllerConfigBase,
                                                                               DirectionalTradingControllerConfigBase]
                                 and (issubclass(member, ControllerConfigBase))), None)
            if not config_class:
                raise InvalidController(f"No configuration class found in the module {controller_name}.")

            loaded_configs.append(config_class(**config_data))

        return loaded_configs

    @field_validator('markets', mode="before")
    @classmethod
    def parse_markets(cls, v) -> Dict[str, Set[str]]:
        if isinstance(v, str):
            return cls.parse_markets_str(v)
        elif isinstance(v, dict):
            return v
        raise ValueError("Invalid type for markets. Expected str or Dict[str, Set[str]]")

    @staticmethod
    def parse_markets_str(v: str) -> Dict[str, Set[str]]:
        markets_dict = {}
        if v.strip():
            exchanges = v.split(':')
            for exchange in exchanges:
                parts = exchange.split('.')
                if len(parts) != 2 or not parts[1]:
                    raise ValueError(f"Invalid market format in segment '{exchange}'. "
                                     "Expected format: 'exchange.tp1,tp2'")
                exchange_name, trading_pairs = parts
                markets_dict[exchange_name] = set(trading_pairs.split(','))
        return markets_dict

    @field_validator('candles_config', mode="before")
    @classmethod
    def parse_candles_config(cls, v) -> List[CandlesConfig]:
        if isinstance(v, str):
            return cls.parse_candles_config_str(v)
        elif isinstance(v, list):
            return v
        raise ValueError("Invalid type for candles_config. Expected str or List[CandlesConfig]")

    @staticmethod
    def parse_candles_config_str(v: str) -> List[CandlesConfig]:
        configs = []
        if v.strip():
            entries = v.split(':')
            for entry in entries:
                parts = entry.split('.')
                if len(parts) != 4:
                    raise ValueError(f"Invalid candles config format in segment '{entry}'. "
                                     "Expected format: 'exchange.tradingpair.interval.maxrecords'")
                connector, trading_pair, interval, max_records_str = parts
                try:
                    max_records = int(max_records_str)
                except ValueError:
                    raise ValueError(f"Invalid max_records value '{max_records_str}' in segment '{entry}'. "
                                     "max_records should be an integer.")
                config = CandlesConfig(
                    connector=connector,
                    trading_pair=trading_pair,
                    interval=interval,
                    max_records=max_records
                )
                configs.append(config)
        return configs


class StrategyV2Base(ScriptStrategyBase):
    """
    V2StrategyBase is a base class for strategies that use the new smart components architecture.
    """
    markets: Dict[str, Set[str]]
    _last_config_update_ts: float = 0
    closed_executors_buffer: int = 100
    max_executors_close_attempts: int = 10
    config_update_interval: int = 10

    @classmethod
    def init_markets(cls, config: StrategyV2ConfigBase):
        """
        Initialize the markets that the strategy is going to use. This method is called when the strategy is created in
        the start command. Can be overridden to implement custom behavior.
        """
        markets = MarketDict(config.markets)
        controllers_configs = config.load_controller_configs()
        for controller_config in controllers_configs:
            markets = controller_config.update_markets(markets)
        cls.markets = markets

    def __init__(self, connectors: Dict[str, ConnectorBase], config: Optional[StrategyV2ConfigBase] = None):
        super().__init__(connectors, config)
        self.config = config

        # Initialize empty dictionaries to hold controllers and unified controller reports
        self.controllers: Dict[str, ControllerBase] = {}
        self.controller_reports: Dict[str, Dict] = {}

        # Initialize the market data provider and executor orchestrator
        self.market_data_provider = MarketDataProvider(connectors)
        self.market_data_provider.initialize_candles_feed_list(config.candles_config)

        # Initialize the controllers
        self.actions_queue = asyncio.Queue()
        self.listen_to_executor_actions_task: asyncio.Task = asyncio.create_task(self.listen_to_executor_actions())
        self.initialize_controllers()
        self._is_stop_triggered = False

        # Collect initial positions from all controller configs
        self.executor_orchestrator = ExecutorOrchestrator(
            strategy=self,
            initial_positions_by_controller=self._collect_initial_positions()
        )
        self.mqtt_enabled = False
        self._pub: Optional[ETopicPublisher] = None

    def start(self, clock: Clock, timestamp: float) -> None:
        """
        Start the strategy.
        :param clock: Clock to use.
        :param timestamp: Current time.
        """
        self._last_timestamp = timestamp
        self.apply_initial_setting()
        # Check if MQTT is enabled at runtime
        from hummingbot.client.hummingbot_application import HummingbotApplication
        if HummingbotApplication.main_application()._mqtt is not None:
            self.mqtt_enabled = True
            self._pub = ETopicPublisher("performance", use_bot_prefix=True)

        # Start controllers
        for controller in self.controllers.values():
            controller.start()

    def apply_initial_setting(self):
        """
        Apply initial settings for the strategy, such as setting position mode and leverage for all connectors.
        """
        pass

    def _collect_initial_positions(self) -> Dict[str, List]:
        """
        Collect initial positions from all controller configurations.
        Returns a dictionary mapping controller_id -> list of InitialPositionConfig.
        """
        if not self.config:
            return {}

        initial_positions_by_controller = {}
        try:
            controllers_configs = self.config.load_controller_configs()
            for controller_config in controllers_configs:
                if hasattr(controller_config, 'initial_positions') and controller_config.initial_positions:
                    initial_positions_by_controller[controller_config.id] = controller_config.initial_positions
        except Exception as e:
            self.logger().error(f"Error collecting initial positions: {e}", exc_info=True)

        return initial_positions_by_controller

    def initialize_controllers(self):
        """
        Initialize the controllers based on the provided configuration.
        """
        controllers_configs = self.config.load_controller_configs()
        for controller_config in controllers_configs:
            self.add_controller(controller_config)
            MarketsRecorder.get_instance().store_controller_config(controller_config)

    def add_controller(self, config: ControllerConfigBase):
        try:
            controller = config.get_controller_class()(config, self.market_data_provider, self.actions_queue)
            self.controllers[config.id] = controller
        except Exception as e:
            self.logger().error(f"Error adding controller: {e}", exc_info=True)

    def update_controllers_configs(self):
        """
        Update the controllers configurations based on the provided configuration.
        """
        if self._last_config_update_ts + self.config_update_interval < self.current_timestamp:
            self._last_config_update_ts = self.current_timestamp
            controllers_configs = self.config.load_controller_configs()
            for controller_config in controllers_configs:
                if controller_config.id in self.controllers:
                    self.controllers[controller_config.id].update_config(controller_config)
                else:
                    self.add_controller(controller_config)

    async def listen_to_executor_actions(self):
        """
        Asynchronously listen to actions from the controllers and execute them.
        """
        while True:
            try:
                actions = await self.actions_queue.get()
                self.executor_orchestrator.execute_actions(actions)
                self.update_executors_info()
                controller_id = actions[0].controller_id
                controller = self.controllers.get(controller_id)
                controller.executors_info = self.get_executors_by_controller(controller_id)
                controller.executors_update_event.set()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(f"Error executing action: {e}", exc_info=True)

    def update_executors_info(self):
        """
        Update the unified controller reports and publish the updates to the active controllers.
        """
        try:
            # Get all reports in a single call and store them
            self.controller_reports = self.executor_orchestrator.get_all_reports()

            # Update each controller with its specific data
            for controller_id, controller in self.controllers.items():
                controller_report = self.controller_reports.get(controller_id, {})
                controller.executors_info = controller_report.get("executors", [])
                controller.positions_held = controller_report.get("positions", [])
                controller.performance_report = controller_report.get("performance", [])
                controller.executors_update_event.set()
        except Exception as e:
            self.logger().error(f"Error updating controller reports: {e}", exc_info=True)

    @staticmethod
    def is_perpetual(connector: str) -> bool:
        return "perpetual" in connector

    async def on_stop(self):
        self._is_stop_triggered = True
        self.listen_to_executor_actions_task.cancel()
        await self.executor_orchestrator.stop(self.max_executors_close_attempts)
        for controller in self.controllers.values():
            controller.stop()
        self.market_data_provider.stop()
        self.executor_orchestrator.store_all_executors()
        if self.mqtt_enabled:
            self._pub({controller_id: {} for controller_id in self.controllers.keys()})
            self._pub = None

    def on_tick(self):
        self.update_executors_info()
        self.update_controllers_configs()
        if self.market_data_provider.ready and not self._is_stop_triggered:
            executor_actions: List[ExecutorAction] = self.determine_executor_actions()
            for action in executor_actions:
                self.executor_orchestrator.execute_action(action)

    def determine_executor_actions(self) -> List[ExecutorAction]:
        """
        Determine actions based on the provided executor handler report.
        """
        actions = []
        actions.extend(self.create_actions_proposal())
        actions.extend(self.stop_actions_proposal())
        actions.extend(self.store_actions_proposal())
        return actions

    def create_actions_proposal(self) -> List[CreateExecutorAction]:
        """
        Create actions proposal based on the current state of the executors.
        """
        raise NotImplementedError

    def stop_actions_proposal(self) -> List[StopExecutorAction]:
        """
        Create a list of actions to stop the executors based on order refresh and early stop conditions.
        """
        raise NotImplementedError

    def store_actions_proposal(self) -> List[StoreExecutorAction]:
        """
        Create a list of actions to store the executors that have been stopped.
        """
        potential_executors_to_store = self.filter_executors(
            executors=self.get_all_executors(),
            filter_func=lambda x: x.is_done)
        sorted_executors = sorted(potential_executors_to_store, key=lambda x: x.timestamp, reverse=True)
        if len(sorted_executors) > self.closed_executors_buffer:
            return [StoreExecutorAction(executor_id=executor.id, controller_id=executor.controller_id) for executor in
                    sorted_executors[self.closed_executors_buffer:]]
        return []

    def get_executors_by_controller(self, controller_id: str) -> List[ExecutorInfo]:
        """Get executors for a specific controller from the unified reports."""
        return self.controller_reports.get(controller_id, {}).get("executors", [])

    def get_all_executors(self) -> List[ExecutorInfo]:
        """Get all executors from all controllers."""
        return [executor for executors_list in [report.get("executors", []) for report in self.controller_reports.values()] for executor in executors_list]

    def get_positions_by_controller(self, controller_id: str) -> List[PositionSummary]:
        """Get positions for a specific controller from the unified reports."""
        return self.controller_reports.get(controller_id, {}).get("positions", [])

    def get_performance_report(self, controller_id: str):
        """Get performance report for a specific controller."""
        return self.controller_reports.get(controller_id, {}).get("performance")

    def set_leverage(self, connector: str, trading_pair: str, leverage: int):
        self.connectors[connector].set_leverage(trading_pair, leverage)

    def set_position_mode(self, connector: str, position_mode: PositionMode):
        self.connectors[connector].set_position_mode(position_mode)

    @staticmethod
    def filter_executors(executors: List[ExecutorInfo], filter_func: Callable[[ExecutorInfo], bool]) -> List[ExecutorInfo]:
        return [executor for executor in executors if filter_func(executor)]

    @staticmethod
    def executors_info_to_df(executors_info: List[ExecutorInfo]) -> pd.DataFrame:
        """
        Convert a list of executor handler info to a dataframe.
        """
        df = pd.DataFrame([ei.to_dict() for ei in executors_info])
        # Convert the enum values to integers
        df['status'] = df['status'].apply(lambda x: x.value)

        # Sort the DataFrame
        df.sort_values(by='status', ascending=True, inplace=True)

        # Convert back to enums for display
        df['status'] = df['status'].apply(RunnableStatus)
        return df

    def format_status(self) -> str:
        if not self.ready_to_trade:
            return "Market connectors are not ready."

        lines = []
        warning_lines = []
        warning_lines.extend(self.network_warning(self.get_market_trading_pair_tuples()))

        # Basic account info
        balance_df = self.get_balance_df()
        lines.extend(["", "  Balances:"] + ["    " + line for line in balance_df.to_string(index=False).split("\n")])

        try:
            df = self.active_orders_df()
            lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
        except ValueError:
            lines.extend(["", "  No active maker orders."])

        # Controller sections
        performance_data = []

        for controller_id, controller in self.controllers.items():
            lines.append(f"\n{'=' * 60}")
            lines.append(f"Controller: {controller_id}")
            lines.append(f"{'=' * 60}")

            # Controller status
            lines.extend(controller.to_format_status())

            # Last 6 executors table
            executors_list = self.get_executors_by_controller(controller_id)
            if executors_list:
                lines.append("\n  Recent Executors (Last 6):")
                # Sort by timestamp and take last 6
                recent_executors = sorted(executors_list, key=lambda x: x.timestamp, reverse=True)[:6]
                executors_df = self.executors_info_to_df(recent_executors)
                if not executors_df.empty:
                    executors_df["age"] = self.current_timestamp - executors_df["timestamp"]
                    executor_columns = ["type", "side", "status", "net_pnl_pct", "net_pnl_quote",
                                        "filled_amount_quote", "is_trading", "close_type", "age"]
                    available_columns = [col for col in executor_columns if col in executors_df.columns]
                    lines.append(format_df_for_printout(executors_df[available_columns],
                                                        table_format="psql", index=False))
            else:
                lines.append("  No executors found.")

            # Positions table
            positions = self.get_positions_by_controller(controller_id)
            if positions:
                lines.append("\n  Positions Held:")
                positions_data = []
                for pos in positions:
                    positions_data.append({
                        "Connector": pos.connector_name,
                        "Trading Pair": pos.trading_pair,
                        "Side": pos.side.name,
                        "Amount": f"{pos.amount:.4f}",
                        "Value (USD)": f"${pos.amount * pos.breakeven_price:.2f}",
                        "Breakeven Price": f"{pos.breakeven_price:.6f}",
                        "Unrealized PnL": f"${pos.unrealized_pnl_quote:+.2f}",
                        "Realized PnL": f"${pos.realized_pnl_quote:+.2f}",
                        "Fees": f"${pos.cum_fees_quote:.2f}"
                    })
                positions_df = pd.DataFrame(positions_data)
                lines.append(format_df_for_printout(positions_df, table_format="psql", index=False))
            else:
                lines.append("  No positions held.")

            # Collect performance data for summary table
            performance_report = self.get_performance_report(controller_id)
            if performance_report:
                performance_data.append({
                    "Controller": controller_id,
                    "Realized PnL": f"${performance_report.realized_pnl_quote:.2f}",
                    "Unrealized PnL": f"${performance_report.unrealized_pnl_quote:.2f}",
                    "Global PnL": f"${performance_report.global_pnl_quote:.2f}",
                    "Global PnL %": f"{performance_report.global_pnl_pct:.2f}%",
                    "Volume Traded": f"${performance_report.volume_traded:.2f}"
                })

        # Performance summary table
        if performance_data:
            lines.append(f"\n{'=' * 80}")
            lines.append("PERFORMANCE SUMMARY")
            lines.append(f"{'=' * 80}")

            # Calculate global totals
            global_realized = sum(Decimal(p["Realized PnL"].replace("$", "")) for p in performance_data)
            global_unrealized = sum(Decimal(p["Unrealized PnL"].replace("$", "")) for p in performance_data)
            global_total = global_realized + global_unrealized
            global_volume = sum(Decimal(p["Volume Traded"].replace("$", "")) for p in performance_data)
            global_pnl_pct = (global_total / global_volume) * 100 if global_volume > 0 else Decimal(0)

            # Add global row
            performance_data.append({
                "Controller": "GLOBAL TOTAL",
                "Realized PnL": f"${global_realized:.2f}",
                "Unrealized PnL": f"${global_unrealized:.2f}",
                "Global PnL": f"${global_total:.2f}",
                "Global PnL %": f"{global_pnl_pct:.2f}%",
                "Volume Traded": f"${global_volume:.2f}"
            })

            performance_df = pd.DataFrame(performance_data)
            lines.append(format_df_for_printout(performance_df, table_format="psql", index=False))

        return "\n".join(lines)



import logging
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor
from os import listdir, path
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from hummingbot.logger.struct_logger import StructLogger, StructLogRecord

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter as _ClientConfigAdapter

STRUCT_LOGGER_SET = False
DEV_STRATEGY_PREFIX = "dev"
_prefix_path = None

# Do not raise exceptions during log handling
logging.setLogRecordFactory(StructLogRecord)
logging.setLoggerClass(StructLogger)

_shared_executor = None
_data_path = None
_cert_path = None


def root_path() -> Path:
    from os.path import join, realpath
    return Path(realpath(join(__file__, "../../")))


def get_executor() -> ThreadPoolExecutor:
    global _shared_executor
    if _shared_executor is None:
        _shared_executor = ThreadPoolExecutor()
    return _shared_executor


def prefix_path() -> str:
    global _prefix_path
    if _prefix_path is None:
        from os.path import join, realpath
        _prefix_path = realpath(join(__file__, "../../"))
    return _prefix_path


def set_prefix_path(p: str):
    global _prefix_path
    _prefix_path = p


def data_path() -> str:
    global _data_path
    if _data_path is None:
        from os.path import join, realpath
        _data_path = realpath(join(prefix_path(), "data"))

    import os
    if not os.path.exists(_data_path):
        os.makedirs(_data_path)
    return _data_path


def set_data_path(path: str):
    global _data_path
    _data_path = path


_independent_package: Optional[bool] = None


def is_independent_package() -> bool:
    global _independent_package
    import os
    if _independent_package is None:
        _independent_package = not os.path.basename(sys.executable).startswith("python")
    return _independent_package


def check_dev_mode():
    try:
        if is_independent_package():
            return False
        if not path.isdir(".git"):
            return False
        current_branch = subprocess.check_output(["git", "symbolic-ref", "--short", "HEAD"]).decode("utf8").rstrip()
        if current_branch != "master":
            return True
    except Exception:
        return False


def chdir_to_data_directory():
    if not is_independent_package():
        # Do nothing.
        return

    import os

    import appdirs
    app_data_dir: str = appdirs.user_data_dir("Hummingbot", "hummingbot.io")
    os.makedirs(os.path.join(app_data_dir, "logs"), 0o711, exist_ok=True)
    os.makedirs(os.path.join(app_data_dir, "conf"), 0o711, exist_ok=True)
    os.makedirs(os.path.join(app_data_dir, "pmm_scripts"), 0o711, exist_ok=True)
    os.makedirs(os.path.join(app_data_dir, "certs"), 0o711, exist_ok=True)
    os.makedirs(os.path.join(app_data_dir, "scripts"), 0o711, exist_ok=True)
    os.chdir(app_data_dir)
    set_prefix_path(app_data_dir)


def get_logging_conf(conf_filename: str = 'hummingbot_logs.yml'):
    import io
    from os.path import join
    from typing import Dict

    from ruamel.yaml import YAML

    file_path: str = join(prefix_path(), "conf", conf_filename)
    yaml_parser: YAML = YAML()
    if not path.exists(file_path):
        return {}
    with open(file_path) as fd:
        yml_source: str = fd.read()
        io_stream: io.StringIO = io.StringIO(yml_source)
        config_dict: Dict = yaml_parser.load(io_stream)
        return config_dict


def init_logging(conf_filename: str,
                 client_config_map: "_ClientConfigAdapter",
                 override_log_level: Optional[str] = None,
                 strategy_file_path: str = "hummingbot"):
    import io
    import logging.config
    from os.path import join
    from typing import Dict

    import pandas as pd
    from ruamel.yaml import YAML

    from hummingbot.logger.struct_logger import StructLogger, StructLogRecord
    global STRUCT_LOGGER_SET
    if not STRUCT_LOGGER_SET:
        logging.setLogRecordFactory(StructLogRecord)
        logging.setLoggerClass(StructLogger)
        STRUCT_LOGGER_SET = True

    # Do not raise exceptions during log handling
    logging.raiseExceptions = False

    file_path: str = join(prefix_path(), "conf", conf_filename)
    yaml_parser: YAML = YAML()
    with open(file_path) as fd:
        yml_source: str = fd.read()
        yml_source = yml_source.replace("$PROJECT_DIR", prefix_path())
        yml_source = yml_source.replace("$DATETIME", pd.Timestamp.now().strftime("%Y-%m-%d-%H-%M-%S"))
        yml_source = yml_source.replace("$STRATEGY_FILE_PATH", strategy_file_path.replace(".yml", ""))
        io_stream: io.StringIO = io.StringIO(yml_source)
        config_dict: Dict = yaml_parser.load(io_stream)
        if override_log_level is not None and "loggers" in config_dict:
            for logger in config_dict["loggers"]:
                if logger in client_config_map.logger_override_whitelist:
                    config_dict["loggers"][logger]["level"] = override_log_level
        logging.config.dictConfig(config_dict)


def get_strategy_list() -> List[str]:
    """
    Search `hummingbot.strategy` folder for all available strategies
    Automatically hide all strategies that starts with "dev" if on master branch
    """
    try:
        folder = path.realpath(path.join(__file__, "../strategy"))
        # Only include valid directories
        strategies = [d for d in listdir(folder) if path.isdir(path.join(folder, d)) and not d.startswith("__")]
        on_dev_mode = check_dev_mode()
        if not on_dev_mode:
            strategies = [s for s in strategies if not s.startswith(DEV_STRATEGY_PREFIX)]
        return sorted(strategies)
    except Exception as e:
        logging.getLogger().warning(f"Error getting strategy set: {str(e)}")
        return []



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PublicSpotKlineV3Api(_message.Message):
    __slots__ = ("interval", "windowStart", "openingPrice", "closingPrice", "highestPrice", "lowestPrice", "volume", "amount", "windowEnd")
    INTERVAL_FIELD_NUMBER: _ClassVar[int]
    WINDOWSTART_FIELD_NUMBER: _ClassVar[int]
    OPENINGPRICE_FIELD_NUMBER: _ClassVar[int]
    CLOSINGPRICE_FIELD_NUMBER: _ClassVar[int]
    HIGHESTPRICE_FIELD_NUMBER: _ClassVar[int]
    LOWESTPRICE_FIELD_NUMBER: _ClassVar[int]
    VOLUME_FIELD_NUMBER: _ClassVar[int]
    AMOUNT_FIELD_NUMBER: _ClassVar[int]
    WINDOWEND_FIELD_NUMBER: _ClassVar[int]
    interval: str
    windowStart: int
    openingPrice: str
    closingPrice: str
    highestPrice: str
    lowestPrice: str
    volume: str
    amount: str
    windowEnd: int
    def __init__(self, interval: _Optional[str] = ..., windowStart: _Optional[int] = ..., openingPrice: _Optional[str] = ..., closingPrice: _Optional[str] = ..., highestPrice: _Optional[str] = ..., lowestPrice: _Optional[str] = ..., volume: _Optional[str] = ..., amount: _Optional[str] = ..., windowEnd: _Optional[int] = ...) -> None: ...



import math
import time
from typing import Optional

from hummingbot.core.data_type.limit_order import LimitOrder


def order_age(order: LimitOrder, current_time: Optional[float] = None) -> int:
    """
    Get the age of a limit order in second.
    :param order: the order to calculate the age for
    :param current_time: the current time in seconds. If not specified the function will take the machine time
    :return: number of seconds since the order was created until the current time
    """
    now = current_time if current_time is not None and not math.isnan(current_time) else _time()
    return int(now - (order.creation_timestamp / 1e6))


def _time() -> float:
    return time.time()



import logging
from enum import Enum
from os.path import join
from typing import TYPE_CHECKING, Optional

from sqlalchemy import MetaData, create_engine, inspect
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm import Query, Session, sessionmaker
from sqlalchemy.schema import DropConstraint, ForeignKeyConstraint, Table

from hummingbot import data_path
from hummingbot.logger.logger import HummingbotLogger
from hummingbot.model import get_declarative_base
from hummingbot.model.metadata import Metadata as LocalMetadata
from hummingbot.model.transaction_base import TransactionBase

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter


class SQLConnectionType(Enum):
    TRADE_FILLS = 1


class SQLConnectionManager(TransactionBase):
    _scm_logger: Optional[HummingbotLogger] = None
    _scm_trade_fills_instance: Optional["SQLConnectionManager"] = None

    LOCAL_DB_VERSION_KEY = "local_db_version"
    LOCAL_DB_VERSION_VALUE = "20230516"

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._scm_logger is None:
            cls._scm_logger = logging.getLogger(__name__)
        return cls._scm_logger

    @classmethod
    def get_declarative_base(cls):
        return get_declarative_base()

    @classmethod
    def get_trade_fills_instance(
        cls, client_config_map: "ClientConfigAdapter", db_name: Optional[str] = None
    ) -> "SQLConnectionManager":
        if cls._scm_trade_fills_instance is None:
            cls._scm_trade_fills_instance = SQLConnectionManager(
                client_config_map, SQLConnectionType.TRADE_FILLS, db_name=db_name
            )
        elif cls.create_db_path(db_name=db_name) != cls._scm_trade_fills_instance.db_path:
            cls._scm_trade_fills_instance = SQLConnectionManager(
                client_config_map, SQLConnectionType.TRADE_FILLS, db_name=db_name
            )
        return cls._scm_trade_fills_instance

    @classmethod
    def create_db_path(cls, db_path: Optional[str] = None, db_name: Optional[str] = None) -> str:
        if db_path is not None:
            return db_path
        if db_name is not None:
            return join(data_path(), f"{db_name}.sqlite")
        else:
            return join(data_path(), "hummingbot_trades.sqlite")

    def __init__(self,
                 client_config_map: "ClientConfigAdapter",
                 connection_type: SQLConnectionType,
                 db_path: Optional[str] = None,
                 db_name: Optional[str] = None,
                 called_from_migrator = False):
        db_path = self.create_db_path(db_path, db_name)
        self.db_path = db_path

        if connection_type is SQLConnectionType.TRADE_FILLS:
            self._engine: Engine = create_engine(client_config_map.db_mode.get_url(self.db_path))
            self._metadata: MetaData = self.get_declarative_base().metadata
            self._metadata.create_all(self._engine)

            # SQLite does not enforce foreign key constraint, but for others engines, we need to drop it.
            # See: `hummingbot/market/markets_recorder.py`, at line 213.
            with self._engine.begin() as conn:
                inspector = inspect(conn)

                for tname, fkcs in reversed(
                        inspector.get_sorted_table_and_fkc_names()):
                    if fkcs:
                        if not self._engine.dialect.supports_alter:
                            continue
                        for fkc in fkcs:
                            fk_constraint = ForeignKeyConstraint((), (), name=fkc)
                            Table(tname, MetaData(), fk_constraint)
                            conn.execute(DropConstraint(fk_constraint))

        self._session_cls = sessionmaker(bind=self._engine)

        if connection_type is SQLConnectionType.TRADE_FILLS and (not called_from_migrator):
            self.check_and_migrate_db(client_config_map)

    @property
    def engine(self) -> Engine:
        return self._engine

    def get_new_session(self) -> Session:
        return self._session_cls()

    def get_local_db_version(self, session: Session):
        query: Query = (session.query(LocalMetadata)
                        .filter(LocalMetadata.key == self.LOCAL_DB_VERSION_KEY))
        result: Optional[LocalMetadata] = query.one_or_none()
        return result

    def check_and_migrate_db(self, client_config_map: "ClientConfigAdapter"):
        from hummingbot.model.db_migration.migrator import Migrator
        with self.get_new_session() as session:
            with session.begin():
                local_db_version = self.get_local_db_version(session=session)
                if local_db_version is None:
                    version_info: LocalMetadata = LocalMetadata(key=self.LOCAL_DB_VERSION_KEY,
                                                                value=self.LOCAL_DB_VERSION_VALUE)
                    session.add(version_info)
                    session.commit()
                else:
                    # There's no past db version to upgrade from at this moment. So we'll just update the version value
                    # if needed.
                    if local_db_version.value < self.LOCAL_DB_VERSION_VALUE:
                        was_migration_successful = Migrator().migrate_db_to_version(
                            client_config_map, self, int(local_db_version.value), int(self.LOCAL_DB_VERSION_VALUE)
                        )
                        if was_migration_successful:
                            # Cannot use variable local_db_version because reference is not valid
                            # since Migrator changed it
                            self.get_local_db_version(session=session).value = self.LOCAL_DB_VERSION_VALUE



import hmac
import time
from typing import Any, Dict, Optional
from urllib.parse import urlencode

import hummingbot.connector.exchange.bybit.bybit_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class BybitAuth(AuthBase):

    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        return self.add_auth_headers(method=request.method, request=request)

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. Bybit does not use this
        functionality
        """
        return self.generate_ws_auth_message()

    def get_referral_code_headers(self):
        """
        Generates referral headers
        :return: a dictionary of auth headers
        """
        headers = {
            "referer": CONSTANTS.HBOT_BROKER_ID
        }
        return headers

    def add_auth_headers(self, method: str, request: Optional[Dict[str, Any]]):
        """
        Add authentication headers in request object

        :param method: HTTP method (POST, PUT, GET)
        :param request: The request to be configured for authenticated interaction

        :return: request object updated with xauth headers
        """
        ts = str(int(time.time() * 10 ** 3))

        headers = {}
        headers["X-BAPI-TIMESTAMP"] = str(ts)
        headers["X-BAPI-API-KEY"] = self.api_key

        if method.value == "POST":
            signature = self._generate_rest_signature(
                timestamp=ts, method=method, payload=request.data)
        else:
            signature = self._generate_rest_signature(
                timestamp=ts, method=method, payload=request.params)

        headers["X-BAPI-SIGN"] = signature
        headers["X-BAPI-SIGN-TYPE"] = str(CONSTANTS.X_API_SIGN_TYPE)
        headers["X-BAPI-RECV-WINDOW"] = str(CONSTANTS.X_API_RECV_WINDOW)
        request.headers = {**request.headers, **headers} if request.headers is not None else headers
        return request

    def _generate_rest_signature(self, timestamp, method: str, payload: Optional[Dict[str, Any]]) -> str:
        if payload is None:
            payload = {}
        if method == RESTMethod.GET:
            param_str = str(timestamp) + self.api_key + CONSTANTS.X_API_RECV_WINDOW + urlencode(payload)
        elif method == RESTMethod.POST:
            param_str = str(timestamp) + self.api_key + CONSTANTS.X_API_RECV_WINDOW + f"{payload}"
        signature = hmac.new(
            bytes(self.secret_key, "utf-8"),
            param_str.encode("utf-8"),
            digestmod="sha256"
        ).hexdigest()
        return signature

    def _generate_ws_signature(self, expires: int):
        signature = str(hmac.new(
            bytes(self.secret_key, "utf-8"),
            bytes(f"GET/realtime{expires}", "utf-8"),
            digestmod="sha256"
        ).hexdigest())
        return signature

    def generate_ws_auth_message(self):
        """
        Generates the authentication message to start receiving messages from
        the 3 private ws channels
        """
        expires = int((self._time() + 10000) * 1000)
        signature = self._generate_ws_signature(expires)
        auth_message = {
            "op": "auth",
            "args": [self.api_key, expires, signature]
        }
        return auth_message

    def _time(self):
        return time.time()



import dataclasses
import logging
from decimal import Decimal
from enum import Enum
from logging import (
    DEBUG,
    INFO,
    WARNING,
    ERROR,
    CRITICAL
)


from .logger import HummingbotLogger

NETWORK = DEBUG + 6


def log_encoder(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, Enum):
        return str(obj)
    elif dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)
    raise TypeError("Object of type '%s' is not JSON serializable" % type(obj).__name__)


__all__ = [
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "NETWORK",
    "HummingbotLogger",
    "log_encoder"
]
logging.setLoggerClass(HummingbotLogger)
logging.addLevelName(NETWORK, "NETWORK")



#!/usr/bin/env python

from hummingbot.core.data_type.order_book import OrderBook


class OrderBookTrackerEntry:
    def __init__(self, trading_pair: str, timestamp: float, order_book: OrderBook):
        self._trading_pair = trading_pair
        self._timestamp = timestamp
        self._order_book = order_book

    def __repr__(self) -> str:
        return (
            f"OrderBookTrackerEntry(trading_pair='{self._trading_pair}', timestamp='{self._timestamp}', "
            f"order_book='{self._order_book}')"
        )

    @property
    def trading_pair(self) -> str:
        return self._trading_pair

    @property
    def timestamp(self) -> float:
        return self._timestamp

    @property
    def order_book(self) -> OrderBook:
        return self._order_book



from typing import Optional, TypeVar

from pydantic import BaseModel

from hummingbot.strategy_v2.executors.data_types import ExecutorConfigBase

ExecutorConfigType = TypeVar("ExecutorConfigType", bound=ExecutorConfigBase)


class ExecutorAction(BaseModel):
    """
    Base class for bot actions.
    """
    controller_id: Optional[str] = "main"


class CreateExecutorAction(ExecutorAction):
    """
    Action to create an executor.
    """
    executor_config: ExecutorConfigType


class StopExecutorAction(ExecutorAction):
    """
    Action to stop an executor.
    """
    executor_id: str
    keep_position: Optional[bool] = False


class StoreExecutorAction(ExecutorAction):
    """
    Action to store an executor.
    """
    executor_id: str



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api-futures.kucoin.com"
HEALTH_CHECK_ENDPOINT = "/api/v1/timestamp"
CANDLES_ENDPOINT = "/api/v1/kline/query"
SYMBOLS_ENDPOINT = "/api/v1/contracts/active"

HB_TO_KUCOIN_MAP = {
    "BTC": "XBT",
}

PUBLIC_WS_DATA_PATH_URL = "/api/v1/bullet-public"

KLINE_PUSH_WEB_SOCKET_TOPIC = "/contractMarket/limitCandle"

INTERVALS = bidict({
    "1s": "1s",  # Implemented for resampling to 1s from trades in quants-lab
    "1m": "1min",
    "5m": "5min",
    "15m": "15min",
    "30m": "30min",
    "1h": "1hour",
    "2h": "2hour",
    "4h": "4hour",
    "6h": "6hour",
    "8h": "8hour",
    "12h": "12hour",
    "1d": "1day",
})

GRANULARITIES = bidict({
    "1m": 1,  # Up to 24 hours
    "5m": 5,  # Up to 10 days
    "15m": 15,  # Up to 30 days
    "30m": 30,  # Up to 60 days
    "1h": 60,  # Up to 120 days
    "2h": 120,  # Up to 240 days
    "4h": 240,  # Up to 480 days
    "6h": 480,  # Up to 720 days
    "8h": 720,
    "12h": 1440,
    "1d": 10080,
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 500
REQUEST_WEIGHT = "REQUEST_WEIGHT"
MAX_REQUEST = 2000
TIME_INTERVAL = 30


RATE_LIMITS = [
    RateLimit(limit_id=REQUEST_WEIGHT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL),
    RateLimit(limit_id=CANDLES_ENDPOINT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=3)]),
    RateLimit(limit_id=SYMBOLS_ENDPOINT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=3)]),
    RateLimit(limit_id=HEALTH_CHECK_ENDPOINT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=2)]),
    RateLimit(limit_id=PUBLIC_WS_DATA_PATH_URL, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=10)]),
]



# distutils: language=c++

import asyncio
import logging
import time
from typing import List

from hummingbot.core.time_iterator import TimeIterator
from hummingbot.core.time_iterator cimport TimeIterator
from hummingbot.core.clock_mode import ClockMode
from hummingbot.logger import HummingbotLogger

s_logger = None


cdef class Clock:
    @classmethod
    def logger(cls) -> HummingbotLogger:
        global s_logger
        if s_logger is None:
            s_logger = logging.getLogger(__name__)
        return s_logger

    def __init__(self, clock_mode: ClockMode, tick_size: float = 1.0, start_time: float = 0.0, end_time: float = 0.0):
        """
        :param clock_mode: either real time mode or back testing mode
        :param tick_size: time interval of each tick
        :param start_time: (back testing mode only) start of simulation in UNIX timestamp
        :param end_time: (back testing mode only) end of simulation in UNIX timestamp. NaN to simulate to end of data.
        """
        self._clock_mode = clock_mode
        self._tick_size = tick_size
        self._start_time = start_time if clock_mode is ClockMode.BACKTEST else (time.time() // tick_size) * tick_size
        self._end_time = end_time
        self._current_tick = self._start_time
        self._child_iterators = []
        self._current_context = None
        self._started = False

    @property
    def clock_mode(self) -> ClockMode:
        return self._clock_mode

    @property
    def start_time(self) -> float:
        return self._start_time

    @property
    def tick_size(self) -> float:
        return self._tick_size

    @property
    def child_iterators(self) -> List[TimeIterator]:
        return self._child_iterators

    @property
    def current_timestamp(self) -> float:
        return self._current_tick

    def __enter__(self) -> Clock:
        if self._current_context is not None:
            raise EnvironmentError("Clock context is not re-entrant.")
        self._current_context = self._child_iterators.copy()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._current_context is not None:
            for iterator in self._current_context:
                (<TimeIterator>iterator).c_stop(self)
        self._current_context = None

    def add_iterator(self, iterator: TimeIterator):
        if self._current_context is not None:
            self._current_context.append(iterator)
        if self._started:
            (<TimeIterator>iterator).c_start(self, self._current_tick)
        self._child_iterators.append(iterator)

    def remove_iterator(self, iterator: TimeIterator):
        if self._current_context is not None and iterator in self._current_context:
            (<TimeIterator>iterator).c_stop(self)
            self._current_context.remove(iterator)
        self._child_iterators.remove(iterator)

    async def run(self):
        await self.run_til(float("nan"))

    async def run_til(self, timestamp: float):
        cdef:
            TimeIterator child_iterator
            double now = time.time()
            double next_tick_time

        if self._current_context is None:
            raise EnvironmentError("run() and run_til() can only be used within the context of a `with...` statement.")

        self._current_tick = (now // self._tick_size) * self._tick_size
        if not self._started:
            for ci in self._current_context:
                child_iterator = ci
                child_iterator.c_start(self, self._current_tick)
            self._started = True

        try:
            while True:
                now = time.time()
                if now >= timestamp:
                    return

                # Sleep until the next tick
                next_tick_time = ((now // self._tick_size) + 1) * self._tick_size
                await asyncio.sleep(next_tick_time - now)
                self._current_tick = next_tick_time

                # Run through all the child iterators.
                for ci in self._current_context:
                    child_iterator = ci
                    try:
                        child_iterator.c_tick(self._current_tick)
                    except StopIteration:
                        self.logger().error("Stop iteration triggered in real time mode. This is not expected.")
                        return
                    except Exception:
                        self.logger().error("Unexpected error running clock tick.", exc_info=True)
        finally:
            for ci in self._current_context:
                child_iterator = ci
                child_iterator._clock = None

    def backtest_til(self, timestamp: float):
        cdef TimeIterator child_iterator

        if not self._started:
            for ci in self._child_iterators:
                child_iterator = ci
                child_iterator.c_start(self, self._start_time)
            self._started = True

        try:
            while not (self._current_tick >= timestamp):
                self._current_tick += self._tick_size
                for ci in self._child_iterators:
                    child_iterator = ci
                    try:
                        child_iterator.c_tick(self._current_tick)
                    except StopIteration:
                        raise
                    except Exception:
                        self.logger().error("Unexpected error running clock tick.", exc_info=True)
        except StopIteration:
            return
        finally:
            for ci in self._child_iterators:
                child_iterator = ci
                child_iterator._clock = None

    def backtest(self):
        self.backtest_til(self._end_time)



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.dexalot.com/privapi"
HEALTH_CHECK_ENDPOINT = "/trading/environments"
CANDLES_ENDPOINT = "/trading/candlechart"

WSS_URL = "wss://api.dexalot.com/api/ws"

# "M5", "M15",  "M30",  "H1"  "H4",  "D1" only these are supported
INTERVALS = bidict({
    "5m": "M5",
    "15m": "M15",
    "30m": "M30",
    "1h": "H1",
    "4h": "H4",
    "1d": "D1",
})

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000

RATE_LIMITS = [
    RateLimit(CANDLES_ENDPOINT, limit=20000, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=20000, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)])]



import asyncio
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.foxbit import (
    foxbit_constants as CONSTANTS,
    foxbit_utils as utils,
    foxbit_web_utils as web_utils,
)
from hummingbot.connector.exchange.foxbit.foxbit_order_book import (
    FoxbitOrderBook,
    FoxbitOrderBookFields,
    FoxbitTradeFields,
)
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.foxbit.foxbit_exchange import FoxbitExchange


class FoxbitAPIOrderBookDataSource(OrderBookTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None
    _trading_pair_exc_id = {}
    _trading_pair_hb_dict = {}
    _ORDER_BOOK_INTERVAL = 1.0
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'FoxbitExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._trade_messages_queue_key = "trade"
        self._diff_messages_queue_key = "order_book_diff"
        self._domain = domain
        self._api_factory = api_factory
        self._first_update_id = {}
        for trading_pair in self._trading_pairs:
            self._first_update_id[trading_pair] = 0

        self._live_stream_connected = {}

    async def get_new_order_book(self, trading_pair: str) -> OrderBook:
        """
        Creates a local instance of the exchange order book for a particular trading pair

        :param trading_pair: the trading pair for which the order book has to be retrieved

        :return: a local copy of the current order book in the exchange
        """
        await self._load_exchange_instrument_id()
        instrument_id = await self._get_instrument_id_from_trading_pair(trading_pair)
        self._live_stream_connected[instrument_id] = False

        snapshot_msg: OrderBookMessage = await self._order_book_snapshot(trading_pair=trading_pair)
        order_book: OrderBook = self.order_book_create_function()
        order_book.apply_snapshot(snapshot_msg.bids, snapshot_msg.asks, snapshot_msg.update_id)
        return order_book

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """

        instrument_id = await self._get_instrument_id_from_trading_pair(trading_pair)
        wait_count = 0

        while (not (instrument_id in self._live_stream_connected) or self._live_stream_connected[instrument_id] is False) and wait_count < 30:
            self.logger().info("Waiting for real time stream before getting a snapshot")
            await asyncio.sleep(self._ORDER_BOOK_INTERVAL)
            wait_count += 1

        symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair),

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.SNAPSHOT_PATH_URL.format(symbol[0]), domain=self._domain),
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.SNAPSHOT_PATH_URL,
        )

        return data

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            for trading_pair in self._trading_pairs:
                # Subscribe OrderBook
                header = utils.get_ws_message_frame(endpoint=CONSTANTS.WS_SUBSCRIBE_ORDER_BOOK,
                                                    msg_type=CONSTANTS.WS_MESSAGE_FRAME_TYPE["Subscribe"],
                                                    payload={"OMSId": 1, "InstrumentId": await self._get_instrument_id_from_trading_pair(trading_pair), "Depth": CONSTANTS.ORDER_BOOK_DEPTH},)
                subscribe_request: WSJSONRequest = WSJSONRequest(payload=web_utils.format_ws_header(header))
                await ws.send(subscribe_request)

                header = utils.get_ws_message_frame(endpoint=CONSTANTS.WS_SUBSCRIBE_TRADES,
                                                    msg_type=CONSTANTS.WS_MESSAGE_FRAME_TYPE["Subscribe"],
                                                    payload={"InstrumentId": await self._get_instrument_id_from_trading_pair(trading_pair)},)
                subscribe_request: WSJSONRequest = WSJSONRequest(payload=web_utils.format_ws_header(header))
                await ws.send(subscribe_request)

            self.logger().info("Subscribed to public order book channel...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=web_utils.websocket_url(), ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = time.time()
        snapshot_msg: OrderBookMessage = FoxbitOrderBook.snapshot_message_from_exchange(
            snapshot,
            snapshot_timestamp,
            metadata={"trading_pair": trading_pair}
        )
        self._first_update_id[trading_pair] = snapshot['sequence_id']
        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if CONSTANTS.WS_SUBSCRIBE_TRADES or CONSTANTS.WS_TRADE_RESPONSE in raw_message['n']:
            full_msg = eval(raw_message['o'].replace(",false,", ",False,"))
            for msg in full_msg:
                instrument_id = int(msg[FoxbitTradeFields.INSTRUMENTID.value])
                trading_pair = ""

                if instrument_id not in self._trading_pair_hb_dict:
                    trading_pair = await self._get_trading_pair_from_instrument_id(instrument_id)
                else:
                    trading_pair = self._trading_pair_hb_dict[instrument_id]

                trade_message = FoxbitOrderBook.trade_message_from_exchange(
                    msg=msg,
                    metadata={"trading_pair": trading_pair},
                )
                message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if CONSTANTS.WS_ORDER_BOOK_RESPONSE or CONSTANTS.WS_ORDER_STATE in raw_message['n']:
            full_msg = eval(raw_message['o'])
            for msg in full_msg:
                instrument_id = int(msg[FoxbitOrderBookFields.PRODUCTPAIRCODE.value])

                trading_pair = ""

                if instrument_id not in self._trading_pair_hb_dict:
                    trading_pair = await self._get_trading_pair_from_instrument_id(instrument_id)
                else:
                    trading_pair = self._trading_pair_hb_dict[instrument_id]

                order_book_message: OrderBookMessage = FoxbitOrderBook.diff_message_from_exchange(
                    msg=msg,
                    metadata={"trading_pair": trading_pair},
                )
                message_queue.put_nowait(order_book_message)
                self._live_stream_connected[instrument_id] = True

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "o" in event_message:
            event_type = event_message.get("n")
            if event_type == CONSTANTS.WS_SUBSCRIBE_TRADES:
                return self._trade_messages_queue_key
            elif event_type == CONSTANTS.WS_ORDER_BOOK_RESPONSE:
                return self._diff_messages_queue_key
        return channel

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _load_exchange_instrument_id(self):
        for trading_pair in self._trading_pairs:
            instrument_id = int(await self._connector.exchange_instrument_id_associated_to_pair(trading_pair=trading_pair))
            self._trading_pair_exc_id[trading_pair] = instrument_id
            self._trading_pair_hb_dict[instrument_id] = trading_pair

    async def _get_trading_pair_from_instrument_id(self, instrument_id: int) -> str:
        if instrument_id not in self._trading_pair_hb_dict:
            await self._load_exchange_instrument_id()
        return self._trading_pair_hb_dict[instrument_id]

    async def _get_instrument_id_from_trading_pair(self, traiding_pair: str) -> int:
        if traiding_pair not in self._trading_pair_exc_id:
            await self._load_exchange_instrument_id()
        return self._trading_pair_exc_id[traiding_pair]

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to subscribe to
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot subscribe: WebSocket connection not established")
            return False

        try:
            instrument_id = await self._get_instrument_id_from_trading_pair(trading_pair)

            # Subscribe OrderBook
            header = utils.get_ws_message_frame(endpoint=CONSTANTS.WS_SUBSCRIBE_ORDER_BOOK,
                                                msg_type=CONSTANTS.WS_MESSAGE_FRAME_TYPE["Subscribe"],
                                                payload={"OMSId": 1, "InstrumentId": instrument_id, "Depth": CONSTANTS.ORDER_BOOK_DEPTH})
            subscribe_request: WSJSONRequest = WSJSONRequest(payload=web_utils.format_ws_header(header))
            await self._ws_assistant.send(subscribe_request)

            header = utils.get_ws_message_frame(endpoint=CONSTANTS.WS_SUBSCRIBE_TRADES,
                                                msg_type=CONSTANTS.WS_MESSAGE_FRAME_TYPE["Subscribe"],
                                                payload={"InstrumentId": instrument_id})
            subscribe_request: WSJSONRequest = WSJSONRequest(payload=web_utils.format_ws_header(header))
            await self._ws_assistant.send(subscribe_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred subscribing to {trading_pair}...",
                exc_info=True
            )
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot unsubscribe: WebSocket connection not established")
            return False

        try:
            instrument_id = await self._get_instrument_id_from_trading_pair(trading_pair)

            # Unsubscribe OrderBook
            header = utils.get_ws_message_frame(endpoint=CONSTANTS.WS_UNSUBSCRIBE_ORDER_BOOK,
                                                msg_type=CONSTANTS.WS_MESSAGE_FRAME_TYPE["Unsubscribe"],
                                                payload={"OMSId": 1, "InstrumentId": instrument_id})
            unsubscribe_request: WSJSONRequest = WSJSONRequest(payload=web_utils.format_ws_header(header))
            await self._ws_assistant.send(unsubscribe_request)

            header = utils.get_ws_message_frame(endpoint=CONSTANTS.WS_UNSUBSCRIBE_TRADES,
                                                msg_type=CONSTANTS.WS_MESSAGE_FRAME_TYPE["Unsubscribe"],
                                                payload={"InstrumentId": instrument_id})
            unsubscribe_request: WSJSONRequest = WSJSONRequest(payload=web_utils.format_ws_header(header))
            await self._ws_assistant.send(unsubscribe_request)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred unsubscribing from {trading_pair}...",
                exc_info=True
            )
            return False



import time
from typing import Optional

import hummingbot.connector.exchange.vertex.vertex_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def public_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint
    :param path_url: a public REST endpoint
    :param domain: the Vertex domain to connect to ("mainnet" or "testnet"). The default value is "mainnet"

    :return: the full URL to the endpoint
    """
    if "/indexer" in path_url:
        return CONSTANTS.ARCHIVE_INDEXER_URLS[domain]
    else:
        return CONSTANTS.BASE_URLS[domain] + path_url


def private_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided REST endpoint

    :param path_url: a private REST endpoint
    :param domain: the domain to connect to ("main" or "testnet"). The default value is "main"

    :return: the full URL to the endpoint
    """
    return public_rest_url(path_url=path_url, domain=domain)


def build_api_factory(
    throttler: Optional[AsyncThrottler] = None,
    auth: Optional[AuthBase] = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    api_factory = WebAssistantsFactory(throttler=throttler, auth=auth)
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(throttler: AsyncThrottler, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> float:
    server_time = float(time.time())
    return server_time



import time
from typing import Optional

from hummingbot.connector.exchange.injective_v2 import injective_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None, domain: str = CONSTANTS.DEFAULT_DOMAIN
) -> float:
    return _time() * 1e3


def _time() -> float:
    return time.time()



from hummingbot.connector.exchange.paper_trade.paper_trade_exchange cimport PaperTradeExchange

cdef class MockPaperExchange(PaperTradeExchange):
    cdef c_set_balanced_order_book(self,
                                   str trading_pair,
                                   double mid_price,
                                   double min_price,
                                   double max_price,
                                   double price_step_size,
                                   double volume_step_size)



import asyncio
from decimal import Decimal
from typing import Any, Dict, Optional

from hummingbot.connector.gateway.gateway_base import GatewayBase
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.utils import async_ttl_cache
from hummingbot.core.utils.async_utils import safe_ensure_future


class GatewaySwap(GatewayBase):
    """
    Handles swap-specific functionality including price quotes and trade execution.
    Maintains order tracking and wallet interactions in the base class.
    """

    @async_ttl_cache(ttl=5, maxsize=10)
    async def get_quote_price(
            self,
            trading_pair: str,
            is_buy: bool,
            amount: Decimal,
            slippage_pct: Optional[Decimal] = None,
            pool_address: Optional[str] = None
    ) -> Optional[Decimal]:
        """
        Retrieves the volume weighted average price. For an AMM DEX connectors, this is the swap price for a given amount.

        :param trading_pair: The market trading pair
        :param is_buy: True for an intention to buy, False for an intention to sell
        :param amount: The amount required (in base token unit)
        :return: The quote price.
        """
        base, quote = trading_pair.split("-")
        side: TradeType = TradeType.BUY if is_buy else TradeType.SELL

        # Pull the price from gateway.
        try:
            resp: Dict[str, Any] = await self._get_gateway_instance().quote_swap(
                network=self.network,
                connector=self.connector_name,
                base_asset=base,
                quote_asset=quote,
                amount=amount,
                side=side,
                slippage_pct=slippage_pct,
                pool_address=pool_address
            )
            price = resp.get("price", None)
            return Decimal(price) if price is not None else None
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().network(
                f"Error getting quote price for {trading_pair} {side} order for {amount} amount.",
                exc_info=True,
                app_warning_msg=str(e)
            )

    async def get_order_price(
            self,
            trading_pair: str,
            is_buy: bool,
            amount: Decimal,
    ) -> Decimal:
        """
        Retreives the price required for an order of a given amount. For AMM DEX connectors, this equals the quote price.
        """
        return await self.get_quote_price(trading_pair, is_buy, amount)

    def buy(self, trading_pair: str, amount: Decimal, order_type: OrderType, price: Decimal, **kwargs) -> str:
        """
        Buys an amount of base token for a given price (or cheaper).
        :param trading_pair: The market trading pair
        :param amount: The order amount (in base token unit)
        :param order_type: Any order type is fine, not needed for this.
        :param price: The maximum price for the order.
        :param kwargs: Additional parameters like quote_id
        :return: A newly created order id (internal).
        """
        return self.place_order(True, trading_pair, amount, price, **kwargs)

    def sell(self, trading_pair: str, amount: Decimal, order_type: OrderType, price: Decimal, **kwargs) -> str:
        """
        Sells an amount of base token for a given price (or at a higher price).
        :param trading_pair: The market trading pair
        :param amount: The order amount (in base token unit)
        :param order_type: Any order type is fine, not needed for this.
        :param price: The minimum price for the order.
        :param kwargs: Additional parameters like quote_id
        :return: A newly created order id (internal).
        """
        return self.place_order(False, trading_pair, amount, price, **kwargs)

    def place_order(self, is_buy: bool, trading_pair: str, amount: Decimal, price: Decimal, **request_args) -> str:
        """
        Places an order.
        :param is_buy: True for buy order
        :param trading_pair: The market trading pair
        :param amount: The order amount (in base token unit)
        :param price: The minimum price for the order.
        :return: A newly created order id (internal).
        """
        side: TradeType = TradeType.BUY if is_buy else TradeType.SELL
        order_id: str = self.create_market_order_id(side, trading_pair)
        safe_ensure_future(self._create_order(side, order_id, trading_pair, amount, price, **request_args))
        return order_id

    async def _create_order(
            self,
            trade_type: TradeType,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            price: Decimal,
            **kwargs
    ):
        """
        Calls buy or sell API end point to place an order, starts tracking the order and triggers relevant order events.
        :param trade_type: BUY or SELL
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The market to place order
        :param amount: The order amount (in base token value)
        :param price: The order price (TO-DO: add limit_price to Gateway execute-swap schema)
        :param kwargs: Additional parameters like quote_id
        """

        amount = self.quantize_order_amount(trading_pair, amount)
        price = self.quantize_order_price(trading_pair, price)

        base, quote = trading_pair.split("-")
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  price=price,
                                  amount=amount)
        try:
            # Check if we have a quote_id to execute
            quote_id = kwargs.get("quote_id")
            if quote_id:
                # Use execute_quote if we have a quote_id
                order_result: Dict[str, Any] = await self._get_gateway_instance().execute_quote(
                    connector=self.connector_name,
                    quote_id=quote_id,
                    network=self.network,
                    wallet_address=self.address
                )
            else:
                # Use execute_swap for direct swaps without quote
                order_result: Dict[str, Any] = await self._get_gateway_instance().execute_swap(
                    connector=self.connector_name,
                    base_asset=base,
                    quote_asset=quote,
                    side=trade_type,
                    amount=amount,
                    network=self.network,
                    wallet_address=self.address
                )
            transaction_hash: Optional[str] = order_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, order_result)

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, f"submitting {trade_type.name} swap order", e)



import abc

from hummingbot.core.web_assistant.connections.data_types import RESTResponse


class RESTPostProcessorBase(abc.ABC):
    """An interface class that enables functionality injection into the `RESTAssistant`.

    The logic provided by a class implementing this interface is applied to a response
    before it is returned to the caller.
    """

    @abc.abstractmethod
    async def post_process(self, response: RESTResponse) -> RESTResponse:
        ...



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.bitget.com"
WSS_URL = "wss://ws.bitget.com/v2/ws/public"

HEALTH_CHECK_ENDPOINT = "/api/v2/public/time"
CANDLES_ENDPOINT = "/api/v2/spot/market/candles"
WS_CANDLES_ENDPOINT = "candle"
PUBLIC_WS_PING_REQUEST = "ping"

WS_HEARTBEAT_TIME_INTERVAL = 30

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000

INTERVAL_LIMITS_DAYS = {
    "1m": 30,
    "3m": 30,
    "5m": 30,
    "15m": 52,
    "30m": 62,
    "1h": 83,
    "2h": 120,
    "4h": 240,
    "6h": 360
}

INTERVALS = bidict({
    "1m": "1min",
    "3m": "3min",
    "5m": "5min",
    "15m": "15min",
    "30m": "30min",
    "1h": "1h",
    "4h": "4h",
    "6h": "6h",
    "12h": "12h",
    "1d": "1day",
    "3d": "3day",
    "1w": "1week",
    "1M": "1M"
})

WS_INTERVALS = bidict({
    "1m": "1m",
    "3m": "3m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "1H",
    "4h": "4H",
    "6h": "6H",
    "12h": "12H",
    "1d": "1D",
    "3d": "3D",
    "1w": "1W",
    "1M": "1M"
})

RATE_LIMITS = [
    RateLimit(CANDLES_ENDPOINT, limit=20, time_interval=1, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=10, time_interval=1, linked_limits=[LinkedLimitWeightPair("raw", 1)])
]



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.hyperliquid.xyz/info"
HEALTH_CHECK_PAYLOAD = {"type": "spotMeta"}
CANDLES_ENDPOINT = "candleSnapshot"

WSS_URL = "wss://api.hyperliquid.xyz/ws"

INTERVALS = bidict({
    "1m": "1m",
    "3m": "3m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "1h",
    "2h": "2h",
    "4h": "4h",
    "6h": "6h",
    "12h": "12h",
    "1d": "1d",
    "1w": "1w",
    "1M": "1M",
})

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 500

RATE_LIMITS = [
    RateLimit(REST_URL, limit=1200, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)])
]

PING_TIMEOUT = 30.0
PING_PAYLOAD = {"method": "ping"}



from decimal import Decimal
from typing import Dict, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.data_feed.coin_cap_data_feed import CoinCapDataFeed
from hummingbot.logger import HummingbotLogger


class CoinCapRateSource(RateSourceBase):
    _logger: Optional[HummingbotLogger] = None

    def __init__(self, assets_map: Dict[str, str], api_key: str):
        self._coin_cap_data_feed = CoinCapDataFeed(assets_map=assets_map, api_key=api_key)

    @property
    def name(self) -> str:
        return "coin_cap"

    async def start_network(self):
        await self._coin_cap_data_feed.start_network()

    async def stop_network(self):
        await self._coin_cap_data_feed.stop_network()

    async def check_network(self) -> NetworkStatus:
        return await self._coin_cap_data_feed.check_network()

    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        prices = {}

        if quote_token == self._coin_cap_data_feed.universal_quote_token:
            prices = await self._coin_cap_data_feed.get_all_usd_quoted_prices()
        else:
            self.logger().warning(
                "CoinCapRateSource only supports USD as quote token. Please set your global token to USD."
            )

        return prices



import asyncio
from decimal import ROUND_UP, Decimal
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from bidict import bidict

import hummingbot.connector.exchange.bitget.bitget_constants as CONSTANTS
from hummingbot.connector.exchange.bitget import bitget_utils, bitget_web_utils as web_utils
from hummingbot.connector.exchange.bitget.bitget_api_order_book_data_source import BitgetAPIOrderBookDataSource
from hummingbot.connector.exchange.bitget.bitget_api_user_stream_data_source import BitgetAPIUserStreamDataSource
from hummingbot.connector.exchange.bitget.bitget_auth import BitgetAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase, TradeFeeSchema
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

s_decimal_NaN = Decimal("nan")


class BitgetExchange(ExchangePyBase):

    web_utils = web_utils

    def __init__(
        self,
        bitget_api_key: str = None,
        bitget_secret_key: str = None,
        bitget_passphrase: str = None,
        balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
        rate_limits_share_pct: Decimal = Decimal("100"),
        trading_pairs: Optional[List[str]] = None,
        trading_required: bool = True,
    ) -> None:
        self._api_key = bitget_api_key
        self._secret_key = bitget_secret_key
        self._passphrase = bitget_passphrase
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs

        self._expected_market_amounts: Dict[str, Decimal] = {}

        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @property
    def authenticator(self) -> BitgetAuth:
        return BitgetAuth(
            api_key=self._api_key,
            secret_key=self._secret_key,
            passphrase=self._passphrase,
            time_provider=self._time_synchronizer
        )

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return CONSTANTS.DEFAULT_DOMAIN

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.ORDER_ID_MAX_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.PUBLIC_SYMBOLS_ENDPOINT

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.PUBLIC_SYMBOLS_ENDPOINT

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PUBLIC_TIME_ENDPOINT

    @property
    def trading_pairs(self) -> Optional[List[str]]:
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @staticmethod
    def _formatted_error(code: int, message: str) -> str:
        return f"Error: {code} - {message}"

    def supported_order_types(self) -> List[OrderType]:
        return [OrderType.LIMIT, OrderType.MARKET]

    def _is_request_exception_related_to_time_synchronizer(
        self,
        request_exception: Exception
    ) -> bool:
        error_description = str(request_exception)
        ts_error_target_str = "Request timestamp expired"

        return ts_error_target_str in error_description

    def _is_order_not_found_during_status_update_error(
        self,
        status_update_exception: Exception
    ) -> bool:
        # Error example:
        # { "code": "00000", "msg": "success", "requestTime": 1710327684832, "data": [] }

        if isinstance(status_update_exception, IOError):
            return any(
                value in str(status_update_exception)
                for value in CONSTANTS.RET_CODES_ORDER_NOT_EXISTS
            )

        if isinstance(status_update_exception, ValueError):
            return True

        return False

    def _is_order_not_found_during_cancelation_error(
        self,
        cancelation_exception: Exception
    ) -> bool:
        # Error example:
        # { "code": "43001", "msg": "订单不存在", "requestTime": 1710327684832, "data": null }

        if isinstance(cancelation_exception, IOError):
            return any(
                value in str(cancelation_exception)
                for value in CONSTANTS.RET_CODES_ORDER_NOT_EXISTS
            )

        return False

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder) -> bool:
        cancel_order_response = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_ENDPOINT,
            data={
                "symbol": await self.exchange_symbol_associated_to_pair(tracked_order.trading_pair),
                "clientOid": tracked_order.client_order_id
            },
            is_auth_required=True,
        )
        response_code = cancel_order_response["code"]

        if response_code != CONSTANTS.RET_CODE_OK:
            raise IOError(self._formatted_error(
                response_code,
                f"Can't cancel order {order_id}: {cancel_order_response}"
            ))

        self._expected_market_amounts.pop(tracked_order.client_order_id, None)

        return True

    async def _place_order(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Decimal,
        **kwargs,
    ) -> Tuple[str, float]:
        if order_type is OrderType.MARKET and trade_type is TradeType.BUY:
            current_price: Decimal = self.get_price(trading_pair, True)
            step_size = Decimal(self.trading_rules[trading_pair].min_base_amount_increment)
            amount = (amount * current_price).quantize(step_size, rounding=ROUND_UP)
            self._expected_market_amounts[order_id] = amount
        data = {
            "side": CONSTANTS.TRADE_TYPES[trade_type],
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair),
            "size": str(amount),
            "orderType": CONSTANTS.ORDER_TYPES[order_type],
            "force": CONSTANTS.DEFAULT_TIME_IN_FORCE,
            "clientOid": order_id,
        }
        if order_type.is_limit_type():
            data["price"] = str(price)

        create_order_response = await self._api_post(
            path_url=CONSTANTS.PLACE_ORDER_ENDPOINT,
            data=data,
            is_auth_required=True,
            headers={
                "X-CHANNEL-API-CODE": CONSTANTS.API_CODE,
            }
        )
        response_code = create_order_response["code"]

        if response_code != CONSTANTS.RET_CODE_OK:
            raise IOError(self._formatted_error(
                response_code,
                f"Error submitting order {order_id}: {create_order_response}"
            ))

        return str(create_order_response["data"]["orderId"]), self.current_timestamp

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)

        if trading_pair in self._trading_fees:
            fee_schema: TradeFeeSchema = self._trading_fees[trading_pair]
            fee_rate = (
                fee_schema.maker_percent_fee_decimal
                if is_maker
                else fee_schema.taker_percent_fee_decimal
            )
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=fee_schema,
                trade_type=order_side,
                percent=fee_rate,
            )
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    async def _update_trading_fees(self) -> None:
        exchange_info = await self._api_get(
            path_url=self.trading_rules_request_path
        )
        symbol_data = exchange_info["data"]

        for symbol_details in symbol_data:
            if bitget_utils.is_exchange_information_valid(exchange_info=symbol_details):
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(
                    symbol=symbol_details["symbol"]
                )
                self._trading_fees[trading_pair] = TradeFeeSchema(
                    maker_percent_fee_decimal=Decimal(symbol_details["makerFeeRate"]),
                    taker_percent_fee_decimal=Decimal(symbol_details["takerFeeRate"])
                )

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            auth=self._auth
        )

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BitgetAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BitgetAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    async def _update_balances(self) -> None:
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        wallet_balance_response: Dict[str, Union[str, List[Dict[str, Any]]]] = await self._api_get(
            path_url=CONSTANTS.ASSETS_ENDPOINT,
            is_auth_required=True,
        )
        response_code = wallet_balance_response["code"]

        if response_code != CONSTANTS.RET_CODE_OK:
            raise IOError(self._formatted_error(
                response_code,
                f"Error while balance update: {wallet_balance_response}"
            ))

        for balance_data in wallet_balance_response["data"]:
            self._set_account_balances(balance_data)
            remote_asset_names.add(balance_data["coin"])

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            try:
                all_fills_response = await self._request_order_fills(order=order)
                fills_data = all_fills_response.get("data", [])

                for fill_data in fills_data:
                    trade_update = self._parse_trade_update(
                        trade_msg=fill_data,
                        tracked_order=order,
                        source_type="rest"
                    )
                    trade_updates.append(trade_update)
            except IOError as ex:
                if not self._is_request_exception_related_to_time_synchronizer(
                    request_exception=ex
                ):
                    raise
        if len(trade_updates) > 0:
            self.logger().info(
                f"{len(trade_updates)} trades updated for order {order.client_order_id}"
            )

        return trade_updates

    async def _request_order_fills(self, order: InFlightOrder) -> Dict[str, Any]:
        order_fills_response = await self._api_get(
            path_url=CONSTANTS.USER_FILLS_ENDPOINT,
            params={
                "orderId": order.exchange_order_id
            },
            is_auth_required=True,
        )

        return order_fills_response

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        order_info_response = await self._request_order_update(tracked_order=tracked_order)

        order_update = self._create_order_update(
            order=tracked_order,
            order_update_response=order_info_response
        )

        return order_update

    def _create_order_update(
        self, order: InFlightOrder, order_update_response: Dict[str, Any]
    ) -> OrderUpdate:
        updated_order_data = order_update_response["data"]

        if not updated_order_data:
            raise ValueError(f"Can't parse order status data. Data: {updated_order_data}")

        updated_info = updated_order_data[0]

        if (
            order.trade_type is TradeType.BUY
            and order.order_type is OrderType.MARKET
            and order.client_order_id not in self._expected_market_amounts
        ):
            self._expected_market_amounts[order.client_order_id] = Decimal(updated_info["size"])

        new_state = CONSTANTS.STATE_TYPES[updated_info["status"]]
        order_update = OrderUpdate(
            trading_pair=order.trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=new_state,
            client_order_id=order.client_order_id,
            exchange_order_id=order.exchange_order_id,
        )

        return order_update

    async def _request_order_update(self, tracked_order: InFlightOrder) -> Dict[str, Any]:
        order_info_response = await self._api_get(
            path_url=CONSTANTS.ORDER_INFO_ENDPOINT,
            params={
                "clientOid": tracked_order.client_order_id
            },
            is_auth_required=True,
        )

        return order_info_response

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        resp_json = await self._api_get(
            path_url=CONSTANTS.PUBLIC_TICKERS_ENDPOINT,
            params={
                "symbol": await self.exchange_symbol_associated_to_pair(trading_pair)
            },
        )

        return float(resp_json["data"][0]["lastPr"])

    def _parse_trade_update(
        self,
        trade_msg: Dict,
        tracked_order: InFlightOrder,
        source_type: Literal["websocket", "rest"]
    ) -> Optional[TradeUpdate]:
        self.logger().debug(f"Data for {source_type} trade update: {trade_msg}")

        fee_detail = trade_msg["feeDetail"]
        trade_fee_data = fee_detail[0] if isinstance(fee_detail, list) else fee_detail
        fee_amount = abs(Decimal(trade_fee_data["totalFee"]))
        fee_coin = trade_fee_data["feeCoin"]
        side = TradeType.BUY if trade_msg["side"] == "buy" else TradeType.SELL

        fee = TradeFeeBase.new_spot_fee(
            fee_schema=self.trade_fee_schema(),
            trade_type=side,
            flat_fees=[TokenAmount(amount=fee_amount, token=fee_coin)],
        )

        trade_id: str = trade_msg["tradeId"]
        trading_pair = tracked_order.trading_pair
        fill_price = Decimal(trade_msg["priceAvg"])
        base_amount = Decimal(trade_msg["size"])
        quote_amount = Decimal(trade_msg["amount"])

        if (
            tracked_order.trade_type is TradeType.BUY
            and tracked_order.order_type is OrderType.MARKET
        ):
            expected_price = (
                self._expected_market_amounts[tracked_order.client_order_id] / tracked_order.amount
            )
            base_amount = (quote_amount / expected_price).quantize(
                Decimal(self.trading_rules[trading_pair].min_base_amount_increment),
                rounding=ROUND_UP
            )

        trade_update: TradeUpdate = TradeUpdate(
            trade_id=trade_id,
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=str(trade_msg["orderId"]),
            trading_pair=trading_pair,
            fill_timestamp=int(trade_msg["uTime"]) * 1e-3,
            fill_price=fill_price,
            fill_base_amount=base_amount,
            fill_quote_amount=quote_amount,
            fee=fee
        )

        return trade_update

    async def _user_stream_event_listener(self) -> None:
        async for event_message in self._iter_user_event_queue():
            try:
                channel = event_message["arg"]["channel"]
                data = event_message["data"]

                self.logger().debug(f"Channel: {channel} - Data: {data}")

                if channel == CONSTANTS.WS_ORDERS_ENDPOINT:
                    for order_msg in data:
                        self._process_order_event_message(order_msg)
                elif channel == CONSTANTS.WS_FILL_ENDPOINT:
                    for fill_msg in data:
                        self._process_fill_event_message(fill_msg)
                elif channel == CONSTANTS.WS_ACCOUNT_ENDPOINT:
                    for wallet_msg in data:
                        self._set_account_balances(wallet_msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")

    def _process_order_event_message(self, order_msg: Dict[str, Any]) -> None:
        """
        Updates in-flight order and triggers cancellation or failure event if needed.
        :param order_msg: The order event message payload
        """
        order_status = CONSTANTS.STATE_TYPES[order_msg["status"]]
        client_order_id = str(order_msg["clientOid"])
        updatable_order = self._order_tracker.all_updatable_orders.get(client_order_id)

        if updatable_order is not None:
            if (
                updatable_order.trade_type is TradeType.BUY
                and updatable_order.order_type is OrderType.MARKET
                and client_order_id not in self._expected_market_amounts
            ):
                self._expected_market_amounts[client_order_id] = Decimal(order_msg["notional"])

            if order_status is OrderState.PARTIALLY_FILLED:
                side = TradeType.BUY if order_msg["side"] == "buy" else TradeType.SELL
                fee_amount = abs(Decimal(order_msg["fillFee"]))
                fee_coin = order_msg["fillFeeCoin"]

                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=side,
                    flat_fees=[TokenAmount(amount=fee_amount, token=fee_coin)],
                )
                trading_pair = updatable_order.trading_pair
                fill_price = Decimal(order_msg["fillPrice"])
                base_amount = Decimal(order_msg["baseVolume"])
                quote_amount = base_amount * fill_price

                if (
                    updatable_order.trade_type is TradeType.BUY
                    and updatable_order.order_type is OrderType.MARKET
                ):
                    expected_price = Decimal(order_msg["notional"]) / updatable_order.amount
                    base_amount = (quote_amount / expected_price).quantize(
                        Decimal(self.trading_rules[trading_pair].min_base_amount_increment),
                        rounding=ROUND_UP
                    )

                new_trade_update: TradeUpdate = TradeUpdate(
                    trade_id=order_msg["tradeId"],
                    client_order_id=client_order_id,
                    exchange_order_id=updatable_order.exchange_order_id,
                    trading_pair=updatable_order.trading_pair,
                    fill_timestamp=int(order_msg["fillTime"]) * 1e-3,
                    fill_price=fill_price,
                    fill_base_amount=base_amount,
                    fill_quote_amount=quote_amount,
                    fee=fee
                )
                self._order_tracker.process_trade_update(new_trade_update)

            new_order_update: OrderUpdate = OrderUpdate(
                trading_pair=updatable_order.trading_pair,
                update_timestamp=int(order_msg["uTime"]) * 1e-3,
                new_state=order_status,
                client_order_id=client_order_id,
                exchange_order_id=order_msg["orderId"],
            )
            self._order_tracker.process_order_update(new_order_update)

    def _process_fill_event_message(self, fill_msg: Dict[str, Any]) -> None:
        try:
            order_id = str(fill_msg.get("orderId", ""))
            trade_id = str(fill_msg.get("tradeId", ""))
            fillable_order = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(
                order_id
            )

            if not fillable_order:
                self.logger().debug(
                    f"Ignoring fill message for order {order_id}: not in in_flight_orders."
                )
                return

            trade_update = self._parse_trade_update(
                trade_msg=fill_msg,
                tracked_order=fillable_order,
                source_type="websocket"
            )
            if trade_update:
                self._order_tracker.process_trade_update(trade_update)

                self.logger().debug(
                    f"Processed fill event for order {fillable_order.client_order_id}: "
                    f"Trade {trade_id}: {fill_msg.get('size')} at {fill_msg.get('priceAvg')}."
                )
        except Exception as e:
            self.logger().error(f"Error processing fill event: {e}", exc_info=True)

    def _set_account_balances(self, data: Dict[str, Any]) -> None:
        symbol = data["coin"]
        available = Decimal(str(data["available"]))
        frozen = Decimal(str(data["frozen"]))
        self._account_balances[symbol] = frozen + available
        self._account_available_balances[symbol] = available

    def _initialize_trading_pair_symbols_from_exchange_info(
        self,
        exchange_info: Dict[str, List[Dict[str, Any]]]
    ) -> None:
        mapping = bidict()
        for symbol_data in exchange_info["data"]:
            if bitget_utils.is_exchange_information_valid(exchange_info=symbol_data):
                try:
                    exchange_symbol = symbol_data["symbol"]
                    base = symbol_data["baseCoin"]
                    quote = symbol_data["quoteCoin"]
                    trading_pair = combine_to_hb_trading_pair(base, quote)
                    mapping[exchange_symbol] = trading_pair
                except Exception as exception:
                    self.logger().error(
                        f"There was an error parsing a trading pair information ({exception})"
                    )
        self._set_trading_pair_symbol_map(mapping)

    async def _format_trading_rules(
        self,
        exchange_info_dict: Dict[str, List[Dict[str, Any]]]
    ) -> List[TradingRule]:
        trading_rules = []
        for rule in exchange_info_dict["data"]:
            if bitget_utils.is_exchange_information_valid(exchange_info=rule):
                try:
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(
                        symbol=rule["symbol"]
                    )
                    trading_rules.append(
                        TradingRule(
                            trading_pair=trading_pair,
                            min_order_size=Decimal(f"1e-{rule['quantityPrecision']}"),
                            min_price_increment=Decimal(f"1e-{rule['pricePrecision']}"),
                            min_base_amount_increment=Decimal(f"1e-{rule['quantityPrecision']}"),
                            min_quote_amount_increment=Decimal(f"1e-{rule['quotePrecision']}"),
                            min_notional_size=Decimal(rule["minTradeUSDT"]),
                        )
                    )
                except Exception:
                    self.logger().exception(
                        f"Error parsing the trading pair rule: {rule}. Skipping."
                    )
        return trading_rules



import asyncio
import time
from typing import TYPE_CHECKING, Any, Dict, Optional

from xrpl.asyncio.clients import AsyncWebsocketClient
from xrpl.models import Subscribe

from hummingbot.connector.exchange.xrpl import xrpl_constants as CONSTANTS
from hummingbot.connector.exchange.xrpl.xrpl_auth import XRPLAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.xrpl.xrpl_exchange import XrplExchange


class XRPLAPIUserStreamDataSource(UserStreamTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None

    def __init__(self, auth: XRPLAuth, connector: "XrplExchange"):
        super().__init__()
        self._connector = connector
        self._auth = auth
        self._last_recv_time: float = 0

    @property
    def last_recv_time(self) -> float:
        """
        Returns the time of the last received message

        :return: the timestamp of the last received message in seconds
        """
        return self._last_recv_time

    async def listen_for_user_stream(self, output: asyncio.Queue):
        """
        Connects to the user private channel in the exchange using a websocket connection. With the established
        connection listens to all balance events and order updates provided by the exchange, and stores them in the
        output queue

        :param output: the queue to use to store the received messages
        """
        while True:
            listener = None
            client: AsyncWebsocketClient | None = None
            node_url: str | None = None
            try:
                self._connector._node_pool.add_burst_tokens(1)
                subscribe = Subscribe(accounts=[self._auth.get_account()])
                client = await self._get_client()
                node_url = client.url
                async with client as ws_client:
                    if ws_client._websocket is None:
                        continue

                    ws_client._websocket.max_size = CONSTANTS.WEBSOCKET_MAX_SIZE_BYTES
                    ws_client._websocket.ping_interval = 10
                    ws_client._websocket.ping_timeout = CONSTANTS.WEBSOCKET_CONNECTION_TIMEOUT

                    # set up a listener task
                    listener = asyncio.create_task(self.on_message(ws_client, output_queue=output))

                    # subscribe to the ledger
                    await ws_client.send(subscribe)

                    # Wait for listener to complete naturally when connection closes
                    # The on_message async iterator will exit when WebSocket closes
                    # WebSocket ping/pong mechanism handles keep-alive automatically
                    await listener
            except asyncio.CancelledError:
                self.logger().info("User stream listener task has been cancelled. Exiting...")
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(f"The websocket connection was closed ({connection_exception})")
                if node_url is not None:
                    self._connector._node_pool.mark_bad_node(node_url)
            except TimeoutError:
                self.logger().warning("Timeout error occurred while listening to user stream. Retrying...")
                if node_url is not None:
                    self._connector._node_pool.mark_bad_node(node_url)
            except Exception:
                self.logger().exception("Unexpected error while listening to user stream. Retrying...")
            finally:
                if listener is not None:
                    listener.cancel()
                    await listener
                if client is not None:
                    await client.close()

    async def on_message(self, client: AsyncWebsocketClient, output_queue: asyncio.Queue):
        async for message in client:
            self._last_recv_time = time.time()
            await self._process_event_message(event_message=message, queue=output_queue)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        queue.put_nowait(event_message)

    async def _get_client(self) -> AsyncWebsocketClient:
        return await self._connector._get_async_client()



import asyncio
import logging
import time
from collections import defaultdict, deque
from dataclasses import dataclass, field
from enum import Enum
from typing import Deque, Dict, List, Optional, Tuple

import pandas as pd

from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.event.events import OrderBookTradeEvent
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class OrderBookTrackerDataSourceType(Enum):
    REMOTE_API = 2
    EXCHANGE_API = 3


@dataclass
class LatencyStats:
    """
    Tracks latency statistics with rolling window for recent samples.
    All times are in milliseconds.

    Supports sampling to reduce overhead on high-frequency message streams.
    """
    ROLLING_WINDOW_SIZE: int = 100  # Keep last 100 samples for recent average
    SAMPLE_RATE: int = 10  # Record 1 out of every N messages for latency (set to 1 to record all)

    count: int = 0
    total_ms: float = 0.0
    min_ms: float = float('inf')
    max_ms: float = 0.0
    _recent_samples: Deque = field(default_factory=lambda: deque(maxlen=100))
    _sample_counter: int = 0  # Internal counter for sampling

    def record(self, latency_ms: float):
        """
        Record a new latency sample.

        Uses sampling to reduce overhead - only records latency details
        for every SAMPLE_RATE messages, but always updates count.
        """
        self.count += 1
        self._sample_counter += 1

        # Always track min/max (cheap operations)
        if latency_ms < self.min_ms:
            self.min_ms = latency_ms
        if latency_ms > self.max_ms:
            self.max_ms = latency_ms

        # Only record full stats every SAMPLE_RATE messages to reduce overhead
        if self._sample_counter >= self.SAMPLE_RATE:
            self._sample_counter = 0
            self.total_ms += latency_ms * self.SAMPLE_RATE  # Approximate total
            self._recent_samples.append(latency_ms)

    @property
    def avg_ms(self) -> float:
        """All-time average latency."""
        return self.total_ms / self.count if self.count > 0 else 0.0

    @property
    def recent_avg_ms(self) -> float:
        """Average latency over recent samples."""
        return sum(self._recent_samples) / len(self._recent_samples) if self._recent_samples else 0.0

    @property
    def recent_samples_count(self) -> int:
        """Number of samples in the rolling window."""
        return len(self._recent_samples)

    def to_dict(self) -> Dict:
        """Convert to dictionary for serialization."""
        return {
            "count": self.count,
            "total_ms": self.total_ms,
            "min_ms": self.min_ms if self.min_ms != float('inf') else 0.0,
            "max_ms": self.max_ms,
            "avg_ms": self.avg_ms,
            "recent_avg_ms": self.recent_avg_ms,
            "recent_samples_count": self.recent_samples_count,
        }


@dataclass
class OrderBookPairMetrics:
    """Metrics for a single trading pair."""
    trading_pair: str

    # Message counts
    diffs_processed: int = 0
    diffs_rejected: int = 0
    snapshots_processed: int = 0
    trades_processed: int = 0
    trades_rejected: int = 0

    # Timestamps (perf_counter for internal timing)
    last_diff_timestamp: float = 0.0
    last_snapshot_timestamp: float = 0.0
    last_trade_timestamp: float = 0.0
    tracking_start_time: float = 0.0

    # Latency tracking
    diff_processing_latency: LatencyStats = field(default_factory=LatencyStats)
    snapshot_processing_latency: LatencyStats = field(default_factory=LatencyStats)
    trade_processing_latency: LatencyStats = field(default_factory=LatencyStats)

    def messages_per_minute(self, current_time: float) -> Dict[str, float]:
        """Calculate messages per minute rates."""
        elapsed_minutes = (current_time - self.tracking_start_time) / 60.0 if self.tracking_start_time > 0 else 0
        if elapsed_minutes <= 0:
            return {"diffs": 0.0, "snapshots": 0.0, "trades": 0.0, "total": 0.0}

        diffs_per_min = self.diffs_processed / elapsed_minutes
        snapshots_per_min = self.snapshots_processed / elapsed_minutes
        trades_per_min = self.trades_processed / elapsed_minutes

        return {
            "diffs": diffs_per_min,
            "snapshots": snapshots_per_min,
            "trades": trades_per_min,
            "total": diffs_per_min + snapshots_per_min + trades_per_min,
        }

    def to_dict(self, current_time: float) -> Dict:
        """Convert to dictionary for serialization."""
        return {
            "trading_pair": self.trading_pair,
            "diffs_processed": self.diffs_processed,
            "diffs_rejected": self.diffs_rejected,
            "snapshots_processed": self.snapshots_processed,
            "trades_processed": self.trades_processed,
            "trades_rejected": self.trades_rejected,
            "last_diff_timestamp": self.last_diff_timestamp,
            "last_snapshot_timestamp": self.last_snapshot_timestamp,
            "last_trade_timestamp": self.last_trade_timestamp,
            "tracking_start_time": self.tracking_start_time,
            "messages_per_minute": self.messages_per_minute(current_time),
            "diff_latency": self.diff_processing_latency.to_dict(),
            "snapshot_latency": self.snapshot_processing_latency.to_dict(),
            "trade_latency": self.trade_processing_latency.to_dict(),
        }


@dataclass
class OrderBookTrackerMetrics:
    """Aggregate metrics for the entire order book tracker."""

    # Global message counts
    total_diffs_processed: int = 0
    total_diffs_rejected: int = 0
    total_diffs_queued: int = 0  # Messages queued before tracking ready
    total_snapshots_processed: int = 0
    total_snapshots_rejected: int = 0
    total_trades_processed: int = 0
    total_trades_rejected: int = 0

    # Timing
    tracker_start_time: float = 0.0

    # Global latency stats
    diff_processing_latency: LatencyStats = field(default_factory=LatencyStats)
    snapshot_processing_latency: LatencyStats = field(default_factory=LatencyStats)
    trade_processing_latency: LatencyStats = field(default_factory=LatencyStats)

    # Per-pair metrics
    per_pair_metrics: Dict[str, OrderBookPairMetrics] = field(default_factory=dict)

    def get_or_create_pair_metrics(self, trading_pair: str) -> OrderBookPairMetrics:
        """Get or create metrics for a trading pair."""
        if trading_pair not in self.per_pair_metrics:
            self.per_pair_metrics[trading_pair] = OrderBookPairMetrics(
                trading_pair=trading_pair,
                tracking_start_time=time.perf_counter(),
            )
        return self.per_pair_metrics[trading_pair]

    def remove_pair_metrics(self, trading_pair: str):
        """Remove metrics for a trading pair."""
        self.per_pair_metrics.pop(trading_pair, None)

    def messages_per_minute(self, current_time: float) -> Dict[str, float]:
        """Calculate global messages per minute rates."""
        elapsed_minutes = (current_time - self.tracker_start_time) / 60.0 if self.tracker_start_time > 0 else 0
        if elapsed_minutes <= 0:
            return {"diffs": 0.0, "snapshots": 0.0, "trades": 0.0, "total": 0.0}

        diffs_per_min = self.total_diffs_processed / elapsed_minutes
        snapshots_per_min = self.total_snapshots_processed / elapsed_minutes
        trades_per_min = self.total_trades_processed / elapsed_minutes

        return {
            "diffs": diffs_per_min,
            "snapshots": snapshots_per_min,
            "trades": trades_per_min,
            "total": diffs_per_min + snapshots_per_min + trades_per_min,
        }

    def to_dict(self) -> Dict:
        """Convert to dictionary for serialization."""
        current_time = time.perf_counter()
        return {
            "total_diffs_processed": self.total_diffs_processed,
            "total_diffs_rejected": self.total_diffs_rejected,
            "total_diffs_queued": self.total_diffs_queued,
            "total_snapshots_processed": self.total_snapshots_processed,
            "total_snapshots_rejected": self.total_snapshots_rejected,
            "total_trades_processed": self.total_trades_processed,
            "total_trades_rejected": self.total_trades_rejected,
            "tracker_start_time": self.tracker_start_time,
            "uptime_seconds": current_time - self.tracker_start_time if self.tracker_start_time > 0 else 0,
            "messages_per_minute": self.messages_per_minute(current_time),
            "diff_latency": self.diff_processing_latency.to_dict(),
            "snapshot_latency": self.snapshot_processing_latency.to_dict(),
            "trade_latency": self.trade_processing_latency.to_dict(),
            "per_pair_metrics": {
                pair: metrics.to_dict(current_time)
                for pair, metrics in self.per_pair_metrics.items()
            },
        }


class OrderBookTracker:
    PAST_DIFF_WINDOW_SIZE: int = 32
    _obt_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._obt_logger is None:
            cls._obt_logger = logging.getLogger(__name__)
        return cls._obt_logger

    def __init__(self, data_source: OrderBookTrackerDataSource, trading_pairs: List[str], domain: Optional[str] = None):
        self._domain: Optional[str] = domain
        self._data_source: OrderBookTrackerDataSource = data_source
        self._trading_pairs: List[str] = trading_pairs
        self._order_books_initialized: asyncio.Event = asyncio.Event()
        self._tracking_tasks: Dict[str, asyncio.Task] = {}
        self._order_books: Dict[str, OrderBook] = {}
        self._tracking_message_queues: Dict[str, asyncio.Queue] = {}
        self._past_diffs_windows: Dict[str, Deque] = defaultdict(lambda: deque(maxlen=self.PAST_DIFF_WINDOW_SIZE))
        self._order_book_diff_stream: asyncio.Queue = asyncio.Queue()
        self._order_book_snapshot_stream: asyncio.Queue = asyncio.Queue()
        self._order_book_trade_stream: asyncio.Queue = asyncio.Queue()
        self._ev_loop: asyncio.BaseEventLoop = asyncio.get_event_loop()
        self._saved_message_queues: Dict[str, Deque[OrderBookMessage]] = defaultdict(lambda: deque(maxlen=1000))

        self._emit_trade_event_task: Optional[asyncio.Task] = None
        self._init_order_books_task: Optional[asyncio.Task] = None
        self._order_book_diff_listener_task: Optional[asyncio.Task] = None
        self._order_book_trade_listener_task: Optional[asyncio.Task] = None
        self._order_book_snapshot_listener_task: Optional[asyncio.Task] = None
        self._order_book_diff_router_task: Optional[asyncio.Task] = None
        self._order_book_snapshot_router_task: Optional[asyncio.Task] = None
        self._update_last_trade_prices_task: Optional[asyncio.Task] = None
        self._order_book_stream_listener_task: Optional[asyncio.Task] = None

        # Metrics tracking
        self._metrics: OrderBookTrackerMetrics = OrderBookTrackerMetrics()

    @property
    def metrics(self) -> OrderBookTrackerMetrics:
        """Access order book tracker metrics."""
        return self._metrics

    @property
    def data_source(self) -> OrderBookTrackerDataSource:
        return self._data_source

    @property
    def order_books(self) -> Dict[str, OrderBook]:
        return self._order_books

    @property
    def ready(self) -> bool:
        return self._order_books_initialized.is_set()

    @property
    def snapshot(self) -> Dict[str, Tuple[pd.DataFrame, pd.DataFrame]]:
        return {
            trading_pair: order_book.snapshot
            for trading_pair, order_book in self._order_books.items()
        }

    def start(self):
        self.stop()
        self._metrics.tracker_start_time = time.perf_counter()
        self._init_order_books_task = safe_ensure_future(
            self._init_order_books()
        )
        self._emit_trade_event_task = safe_ensure_future(
            self._emit_trade_event_loop()
        )
        self._order_book_diff_listener_task = safe_ensure_future(
            self._data_source.listen_for_order_book_diffs(self._ev_loop, self._order_book_diff_stream)
        )
        self._order_book_trade_listener_task = safe_ensure_future(
            self._data_source.listen_for_trades(self._ev_loop, self._order_book_trade_stream)
        )
        self._order_book_snapshot_listener_task = safe_ensure_future(
            self._data_source.listen_for_order_book_snapshots(self._ev_loop, self._order_book_snapshot_stream)
        )
        self._order_book_stream_listener_task = safe_ensure_future(
            self._data_source.listen_for_subscriptions()
        )
        self._order_book_diff_router_task = safe_ensure_future(
            self._order_book_diff_router()
        )
        self._order_book_snapshot_router_task = safe_ensure_future(
            self._order_book_snapshot_router()
        )
        self._update_last_trade_prices_task = safe_ensure_future(
            self._update_last_trade_prices_loop()
        )

    def stop(self):
        if self._init_order_books_task is not None:
            self._init_order_books_task.cancel()
            self._init_order_books_task = None
        if self._emit_trade_event_task is not None:
            self._emit_trade_event_task.cancel()
            self._emit_trade_event_task = None
        if self._order_book_diff_listener_task is not None:
            self._order_book_diff_listener_task.cancel()
            self._order_book_diff_listener_task = None
        if self._order_book_snapshot_listener_task is not None:
            self._order_book_snapshot_listener_task.cancel()
            self._order_book_snapshot_listener_task = None
        if self._order_book_trade_listener_task is not None:
            self._order_book_trade_listener_task.cancel()
            self._order_book_trade_listener_task = None

        if self._order_book_diff_router_task is not None:
            self._order_book_diff_router_task.cancel()
            self._order_book_diff_router_task = None
        if self._order_book_snapshot_router_task is not None:
            self._order_book_snapshot_router_task.cancel()
            self._order_book_snapshot_router_task = None
        if self._update_last_trade_prices_task is not None:
            self._update_last_trade_prices_task.cancel()
            self._update_last_trade_prices_task = None
        if self._order_book_stream_listener_task is not None:
            self._order_book_stream_listener_task.cancel()
        if len(self._tracking_tasks) > 0:
            for _, task in self._tracking_tasks.items():
                task.cancel()
            self._tracking_tasks.clear()
        self._order_books_initialized.clear()

    async def wait_ready(self):
        await self._order_books_initialized.wait()

    async def _update_last_trade_prices_loop(self):
        '''
        Updates last trade price for all order books through REST API, it is to initiate last_trade_price and as
        fall-back mechanism for when the web socket update channel fails.
        '''
        await self._order_books_initialized.wait()
        while True:
            try:
                outdateds = [t_pair for t_pair, o_book in self._order_books.items()
                             if o_book.last_applied_trade < time.perf_counter() - (60. * 3)
                             and o_book.last_trade_price_rest_updated < time.perf_counter() - 5]
                if outdateds:
                    args = {"trading_pairs": outdateds}
                    if self._domain is not None:
                        args["domain"] = self._domain
                    last_prices = await self._data_source.get_last_traded_prices(**args)
                    for trading_pair, last_price in last_prices.items():
                        self._order_books[trading_pair].last_trade_price = last_price
                        self._order_books[trading_pair].last_trade_price_rest_updated = time.perf_counter()
                else:
                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network("Unexpected error while fetching last trade price.", exc_info=True)
                await asyncio.sleep(30)

    async def _initial_order_book_for_trading_pair(self, trading_pair: str) -> OrderBook:
        return await self._data_source.get_new_order_book(trading_pair)

    async def _init_order_books(self):
        """
        Initialize order books
        """
        for index, trading_pair in enumerate(self._trading_pairs):
            self._order_books[trading_pair] = await self._initial_order_book_for_trading_pair(trading_pair)
            self._tracking_message_queues[trading_pair] = asyncio.Queue()
            self._tracking_tasks[trading_pair] = safe_ensure_future(self._track_single_book(trading_pair))
            self.logger().info(f"Initialized order book for {trading_pair}. "
                               f"{index + 1}/{len(self._trading_pairs)} completed.")
            await self._sleep(delay=1)
        self._order_books_initialized.set()

    async def add_trading_pair(self, trading_pair: str) -> bool:
        """
        Dynamically adds a new trading pair to the order book tracker.

        This method:
        1. Subscribes to the trading pair on the existing WebSocket connection
        2. Fetches the initial order book snapshot
        3. Creates the tracking queue and starts the tracking task
        4. Any messages received before the snapshot (stored in _saved_message_queues)
           will be automatically processed by _track_single_book

        :param trading_pair: the trading pair to add (e.g., "BTC-USDT")
        :return: True if successfully added, False otherwise
        """
        # Check if already tracking this pair
        if trading_pair in self._order_books:
            self.logger().warning(f"Trading pair {trading_pair} is already being tracked")
            return False

        # Wait for initial order books to be ready before adding new ones
        await self._order_books_initialized.wait()

        try:
            self.logger().info(f"Adding trading pair {trading_pair} to order book tracker...")

            # Step 1: Subscribe to WebSocket channels for this pair
            # This ensures we start receiving diff messages immediately
            subscribe_success = await self._data_source.subscribe_to_trading_pair(trading_pair)
            if not subscribe_success:
                self.logger().error(f"Failed to subscribe to {trading_pair} WebSocket channels")
                return False

            # Step 2: Add to internal trading pairs list
            if trading_pair not in self._trading_pairs:
                self._trading_pairs.append(trading_pair)

            # Step 3: Fetch initial snapshot and create order book
            # Note: Diffs received during this time are saved in _saved_message_queues
            self._order_books[trading_pair] = await self._initial_order_book_for_trading_pair(trading_pair)

            # Step 4: Create message queue and start tracking task
            self._tracking_message_queues[trading_pair] = asyncio.Queue()
            self._tracking_tasks[trading_pair] = safe_ensure_future(
                self._track_single_book(trading_pair)
            )

            self.logger().info(f"Successfully added trading pair {trading_pair} to order book tracker")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error adding trading pair {trading_pair}")
            # Clean up partial state
            self._order_books.pop(trading_pair, None)
            self._tracking_message_queues.pop(trading_pair, None)
            if trading_pair in self._tracking_tasks:
                self._tracking_tasks[trading_pair].cancel()
                self._tracking_tasks.pop(trading_pair, None)
            return False

    async def remove_trading_pair(self, trading_pair: str) -> bool:
        """
        Dynamically removes a trading pair from the order book tracker.

        This method:
        1. Cancels the tracking task for the pair
        2. Unsubscribes from WebSocket channels
        3. Cleans up all data structures

        :param trading_pair: the trading pair to remove (e.g., "SOL-USDT")
        :return: True if successfully removed, False otherwise
        """
        # Check if we're tracking this pair
        if trading_pair not in self._order_books:
            self.logger().warning(f"Trading pair {trading_pair} is not being tracked")
            return False

        try:
            self.logger().info(f"Removing trading pair {trading_pair} from order book tracker...")

            # Step 1: Cancel the tracking task
            if trading_pair in self._tracking_tasks:
                self._tracking_tasks[trading_pair].cancel()
                try:
                    await self._tracking_tasks[trading_pair]
                except asyncio.CancelledError:
                    pass
                self._tracking_tasks.pop(trading_pair, None)

            # Step 2: Unsubscribe from WebSocket channels
            unsubscribe_success = await self._data_source.unsubscribe_from_trading_pair(trading_pair)
            if not unsubscribe_success:
                self.logger().warning(f"Failed to unsubscribe from {trading_pair} WebSocket channels")
                # Continue with cleanup anyway

            # Step 3: Clean up data structures
            self._order_books.pop(trading_pair, None)
            self._tracking_message_queues.pop(trading_pair, None)
            self._past_diffs_windows.pop(trading_pair, None)
            self._saved_message_queues.pop(trading_pair, None)

            # Step 4: Clean up metrics for this pair
            self._metrics.remove_pair_metrics(trading_pair)

            # Step 5: Remove from trading pairs list
            if trading_pair in self._trading_pairs:
                self._trading_pairs.remove(trading_pair)

            self.logger().info(f"Successfully removed trading pair {trading_pair} from order book tracker")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error removing trading pair {trading_pair}")
            return False

    async def _order_book_diff_router(self):
        """
        Routes the real-time order book diff messages to the correct order book.
        """
        last_message_timestamp: float = time.time()
        messages_queued: int = 0
        messages_accepted: int = 0
        messages_rejected: int = 0

        # Cache pair_metrics references to avoid repeated dict lookups
        pair_metrics_cache: Dict[str, OrderBookPairMetrics] = {}

        while True:
            try:
                ob_message: OrderBookMessage = await self._order_book_diff_stream.get()
                process_start = time.perf_counter()
                trading_pair: str = ob_message.trading_pair

                if trading_pair not in self._tracking_message_queues:
                    messages_queued += 1
                    self._metrics.total_diffs_queued += 1
                    # Save diff messages received before snapshots are ready
                    self._saved_message_queues[trading_pair].append(ob_message)
                    continue

                message_queue: asyncio.Queue = self._tracking_message_queues[trading_pair]
                order_book: OrderBook = self._order_books[trading_pair]

                # Get or cache pair metrics (single lookup per pair)
                if trading_pair not in pair_metrics_cache:
                    pair_metrics_cache[trading_pair] = self._metrics.get_or_create_pair_metrics(trading_pair)
                pair_metrics = pair_metrics_cache[trading_pair]

                # Check the order book's initial update ID. If it's larger, don't bother.
                if order_book.snapshot_uid > ob_message.update_id:
                    messages_rejected += 1
                    self._metrics.total_diffs_rejected += 1
                    pair_metrics.diffs_rejected += 1
                    continue

                await message_queue.put(ob_message)
                messages_accepted += 1

                # Record metrics (latency uses sampling internally to reduce overhead)
                process_time_ms = (time.perf_counter() - process_start) * 1000
                self._metrics.total_diffs_processed += 1
                self._metrics.diff_processing_latency.record(process_time_ms)
                pair_metrics.diffs_processed += 1
                pair_metrics.last_diff_timestamp = process_start
                pair_metrics.diff_processing_latency.record(process_time_ms)

                # Log some statistics.
                now: float = time.time()
                if int(now / 60.0) > int(last_message_timestamp / 60.0):
                    self.logger().debug(f"Diff messages processed: {messages_accepted}, "
                                        f"rejected: {messages_rejected}, queued: {messages_queued}")
                    messages_accepted = 0
                    messages_rejected = 0
                    messages_queued = 0

                last_message_timestamp = now
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unexpected error routing order book messages.",
                    exc_info=True,
                    app_warning_msg="Unexpected error routing order book messages. Retrying after 5 seconds."
                )
                await asyncio.sleep(5.0)

    async def _order_book_snapshot_router(self):
        """
        Route the real-time order book snapshot messages to the correct order book.
        """
        await self._order_books_initialized.wait()

        # Cache pair_metrics references
        pair_metrics_cache: Dict[str, OrderBookPairMetrics] = {}

        while True:
            try:
                ob_message: OrderBookMessage = await self._order_book_snapshot_stream.get()
                process_start = time.perf_counter()
                trading_pair: str = ob_message.trading_pair

                if trading_pair not in self._tracking_message_queues:
                    self._metrics.total_snapshots_rejected += 1
                    continue

                message_queue: asyncio.Queue = self._tracking_message_queues[trading_pair]
                await message_queue.put(ob_message)

                # Record metrics
                process_time_ms = (time.perf_counter() - process_start) * 1000
                self._metrics.total_snapshots_processed += 1
                self._metrics.snapshot_processing_latency.record(process_time_ms)

                # Get or cache pair metrics
                if trading_pair not in pair_metrics_cache:
                    pair_metrics_cache[trading_pair] = self._metrics.get_or_create_pair_metrics(trading_pair)
                pair_metrics = pair_metrics_cache[trading_pair]
                pair_metrics.snapshots_processed += 1
                pair_metrics.last_snapshot_timestamp = process_start
                pair_metrics.snapshot_processing_latency.record(process_time_ms)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unknown error. Retrying after 5 seconds.", exc_info=True)
                await asyncio.sleep(5.0)

    async def _track_single_book(self, trading_pair: str):
        past_diffs_window = self._past_diffs_windows[trading_pair]

        message_queue: asyncio.Queue = self._tracking_message_queues[trading_pair]
        order_book: OrderBook = self._order_books[trading_pair]
        last_message_timestamp: float = time.time()
        diff_messages_accepted: int = 0

        while True:
            try:
                saved_messages: Deque[OrderBookMessage] = self._saved_message_queues[trading_pair]

                # Process saved messages first if there are any
                if len(saved_messages) > 0:
                    message = saved_messages.popleft()
                else:
                    message = await message_queue.get()

                if message.type is OrderBookMessageType.DIFF:
                    order_book.apply_diffs(message.bids, message.asks, message.update_id)
                    past_diffs_window.append(message)
                    diff_messages_accepted += 1

                    # Output some statistics periodically.
                    now: float = time.time()
                    if int(now / 60.0) > int(last_message_timestamp / 60.0):
                        self.logger().debug(f"Processed {diff_messages_accepted} order book diffs for {trading_pair}.")
                        diff_messages_accepted = 0
                    last_message_timestamp = now
                elif message.type is OrderBookMessageType.SNAPSHOT:
                    past_diffs: List[OrderBookMessage] = list(past_diffs_window)
                    order_book.restore_from_snapshot_and_diffs(message, past_diffs)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    f"Unexpected error tracking order book for {trading_pair}.",
                    exc_info=True,
                    app_warning_msg="Unexpected error tracking order book. Retrying after 5 seconds."
                )
                await asyncio.sleep(5.0)

    async def _emit_trade_event_loop(self):
        last_message_timestamp: float = time.time()
        messages_accepted: int = 0
        messages_rejected: int = 0
        await self._order_books_initialized.wait()

        # Cache pair_metrics references
        pair_metrics_cache: Dict[str, OrderBookPairMetrics] = {}

        while True:
            try:
                trade_message: OrderBookMessage = await self._order_book_trade_stream.get()
                process_start = time.perf_counter()
                trading_pair: str = trade_message.trading_pair

                if trading_pair not in self._order_books:
                    messages_rejected += 1
                    self._metrics.total_trades_rejected += 1
                    continue

                order_book: OrderBook = self._order_books[trading_pair]
                order_book.apply_trade(OrderBookTradeEvent(
                    trading_pair=trade_message.trading_pair,
                    timestamp=trade_message.timestamp,
                    price=float(trade_message.content["price"]),
                    amount=float(trade_message.content["amount"]),
                    trade_id=trade_message.trade_id,
                    type=TradeType.SELL if
                    trade_message.content["trade_type"] == float(TradeType.SELL.value) else TradeType.BUY
                ))

                messages_accepted += 1

                # Record metrics
                process_time_ms = (time.perf_counter() - process_start) * 1000
                self._metrics.total_trades_processed += 1
                self._metrics.trade_processing_latency.record(process_time_ms)

                # Get or cache pair metrics
                if trading_pair not in pair_metrics_cache:
                    pair_metrics_cache[trading_pair] = self._metrics.get_or_create_pair_metrics(trading_pair)
                pair_metrics = pair_metrics_cache[trading_pair]
                pair_metrics.trades_processed += 1
                pair_metrics.last_trade_timestamp = process_start
                pair_metrics.trade_processing_latency.record(process_time_ms)

                # Log some statistics.
                now: float = time.time()
                if int(now / 60.0) > int(last_message_timestamp / 60.0):
                    self.logger().debug(f"Trade messages processed: {messages_accepted}, rejected: {messages_rejected}")
                    messages_accepted = 0
                    messages_rejected = 0

                last_message_timestamp = now
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unexpected error routing order book messages.",
                    exc_info=True,
                    app_warning_msg="Unexpected error routing order book messages. Retrying after 5 seconds."
                )
                await asyncio.sleep(5.0)

    @staticmethod
    async def _sleep(delay: float):
        await asyncio.sleep(delay=delay)



from .balance_command import BalanceCommand
from .config_command import ConfigCommand
from .connect_command import ConnectCommand
from .create_command import CreateCommand
from .exit_command import ExitCommand
from .export_command import ExportCommand
from .gateway_approve_command import GatewayApproveCommand
from .gateway_command import GatewayCommand
from .gateway_lp_command import GatewayLPCommand
from .gateway_pool_command import GatewayPoolCommand
from .gateway_swap_command import GatewaySwapCommand
from .gateway_token_command import GatewayTokenCommand
from .help_command import HelpCommand
from .history_command import HistoryCommand
from .import_command import ImportCommand
from .mqtt_command import MQTTCommand
from .order_book_command import OrderBookCommand
from .rate_command import RateCommand
from .silly_commands import SillyCommands
from .start_command import StartCommand
from .status_command import StatusCommand
from .stop_command import StopCommand
from .ticker_command import TickerCommand

__all__ = [
    BalanceCommand,
    ConfigCommand,
    ConnectCommand,
    CreateCommand,
    ExitCommand,
    ExportCommand,
    GatewayApproveCommand,
    GatewayCommand,
    GatewayLPCommand,
    GatewayPoolCommand,
    GatewaySwapCommand,
    GatewayTokenCommand,
    HelpCommand,
    HistoryCommand,
    ImportCommand,
    OrderBookCommand,
    RateCommand,
    SillyCommands,
    StartCommand,
    StatusCommand,
    StopCommand,
    TickerCommand,
    MQTTCommand,
]



import asyncio
import importlib
import inspect
from decimal import Decimal
from typing import TYPE_CHECKING, Callable, List, Optional

from pydantic import ConfigDict, Field, field_validator

from hummingbot.client.config.config_data_types import BaseClientModel
from hummingbot.core.data_type.common import MarketDict
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.data_feed.market_data_provider import MarketDataProvider
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executor_actions import ExecutorAction
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo, PerformanceReport
from hummingbot.strategy_v2.models.position_config import InitialPositionConfig
from hummingbot.strategy_v2.runnable_base import RunnableBase
from hummingbot.strategy_v2.utils.common import generate_unique_id

if TYPE_CHECKING:
    from hummingbot.strategy_v2.executors.data_types import PositionSummary


class ControllerConfigBase(BaseClientModel):
    """
    This class represents the base configuration for a controller in the Hummingbot trading bot.
    It inherits from the Pydantic BaseModel and includes several fields that are used to configure a controller.

    Attributes:
        id (str): A unique identifier for the controller. If not provided, it will be automatically generated.
        controller_name (str): The name of the trading strategy that the controller will use.
        candles_config (List[CandlesConfig]): A list of configurations for the candles data feed.
    """
    id: str = Field(default=None,)
    controller_name: str
    controller_type: str = "generic"
    total_amount_quote: Decimal = Field(
        default=Decimal("100"),
        json_schema_extra={
            "prompt": "Enter the total amount in quote asset to use for trading (e.g., 1000): ",
            "prompt_on_new": True,
            "is_updatable": True
        }
    )
    manual_kill_switch: bool = Field(default=False, json_schema_extra={"is_updatable": True})
    candles_config: List[CandlesConfig] = Field(
        default=[],
        json_schema_extra={"is_updatable": True})
    initial_positions: List[InitialPositionConfig] = Field(
        default=[],
        json_schema_extra={
            "prompt": "Enter initial positions as a list of InitialPositionConfig objects: ",
            "prompt_on_new": False,
            "is_updatable": False
        })
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator('id', mode="before")
    @classmethod
    def set_id(cls, v):
        if v is None or v.strip() == "":
            return generate_unique_id()
        return v

    @field_validator('candles_config', mode="before")
    @classmethod
    def parse_candles_config(cls, v) -> List[CandlesConfig]:
        if isinstance(v, str):
            return cls.parse_candles_config_str(v)
        elif isinstance(v, list):
            return v
        raise ValueError("Invalid type for candles_config. Expected str or List[CandlesConfig]")

    @field_validator('initial_positions', mode="before")
    @classmethod
    def parse_initial_positions(cls, v) -> List[InitialPositionConfig]:
        if isinstance(v, list):
            return v
        raise ValueError("Invalid type for initial_positions. Expected List[InitialPositionConfig]")

    @staticmethod
    def parse_candles_config_str(v: str) -> List[CandlesConfig]:
        configs = []
        if v.strip():
            entries = v.split(':')
            for entry in entries:
                parts = entry.split('.')
                if len(parts) != 4:
                    raise ValueError(f"Invalid candles config format in segment '{entry}'. "
                                     "Expected format: 'exchange.tradingpair.interval.maxrecords'")
                connector, trading_pair, interval, max_records_str = parts
                try:
                    max_records = int(max_records_str)
                except ValueError:
                    raise ValueError(f"Invalid max_records value '{max_records_str}' in segment '{entry}'. "
                                     "max_records should be an integer.")
                config = CandlesConfig(
                    connector=connector,
                    trading_pair=trading_pair,
                    interval=interval,
                    max_records=max_records
                )
                configs.append(config)
        return configs

    def update_markets(self, markets: MarketDict) -> MarketDict:
        """
        Update the markets dict of the script from the config.
        """
        return markets

    def get_controller_class(self):
        """
        Dynamically load and return the controller class based on the controller configuration.
        """
        try:
            module = importlib.import_module(self.__module__)
            base_classes = ["ControllerBase", "MarketMakingControllerBase", "DirectionalTradingControllerBase"]
            for name, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and issubclass(obj, ControllerBase) and obj.__name__ not in base_classes:
                    return obj
        except ImportError as e:
            raise ImportError(f"Could not import the module: {self.__module__}. Error: {str(e)}")

        raise ValueError(f"No valid controller class found for module: {self.__module__}")


class ControllerBase(RunnableBase):
    """
    Base class for controllers.
    """

    def __init__(self, config: ControllerConfigBase, market_data_provider: MarketDataProvider,
                 actions_queue: asyncio.Queue, update_interval: float = 1.0):
        super().__init__(update_interval=update_interval)
        self.config = config
        self.executors_info: List[ExecutorInfo] = []
        self.positions_held: List[PositionSummary] = []
        self.performance_report: Optional[PerformanceReport] = None
        self.market_data_provider: MarketDataProvider = market_data_provider
        self.actions_queue: asyncio.Queue = actions_queue
        self.processed_data = {}
        self.executors_update_event = asyncio.Event()
        self.executors_info_queue = asyncio.Queue()

    def start(self):
        """
        Allow controllers to be restarted after being stopped.
        """
        if self._status != RunnableStatus.RUNNING:
            self.terminated.clear()
            self._status = RunnableStatus.RUNNING
            self.executors_update_event.set()
            safe_ensure_future(self.control_loop())
        self.initialize_candles()

    def initialize_candles(self):
        for candles_config in self.config.candles_config:
            self.market_data_provider.initialize_candles_feed(candles_config)

    def update_config(self, new_config: ControllerConfigBase):
        """
        Update the controller configuration. With the variables that in the client_data have the is_updatable flag set
        to True. This will be only available for those variables that don't interrupt the bot operation.
        """
        for name, field_info in self.config.__class__.model_fields.items():
            json_schema_extra = field_info.json_schema_extra or {}
            if json_schema_extra.get("is_updatable", False):
                setattr(self.config, name, getattr(new_config, name))

    async def control_task(self):
        if self.market_data_provider.ready and self.executors_update_event.is_set():
            await self.update_processed_data()
            executor_actions: List[ExecutorAction] = self.determine_executor_actions()
            if len(executor_actions) > 0:
                self.logger().debug(f"Sending actions: {executor_actions}")
                await self.send_actions(executor_actions)

    async def send_actions(self, executor_actions: List[ExecutorAction]):
        if len(executor_actions) > 0:
            await self.actions_queue.put(executor_actions)
            self.executors_update_event.clear()  # Clear the event after sending the actions

    @staticmethod
    def filter_executors(executors: List[ExecutorInfo], filter_func: Callable[[ExecutorInfo], bool]) -> List[ExecutorInfo]:
        return [executor for executor in executors if filter_func(executor)]

    async def update_processed_data(self):
        """
        This method should be overridden by the derived classes to implement the logic to update the market data
        used by the controller. And should update the local market data collection to be used by the controller to
        take decisions.
        """
        raise NotImplementedError

    def determine_executor_actions(self) -> List[ExecutorAction]:
        """
        This method should be overridden by the derived classes to implement the logic to determine the actions
        that the executors should take.
        """
        raise NotImplementedError

    def to_format_status(self) -> List[str]:
        """
        This method should be overridden by the derived classes to implement the logic to format the status of the
        controller to be displayed in the UI.
        """
        return []

    def get_custom_info(self) -> dict:
        """
        Override this method to provide custom controller-specific information that will be
        published alongside the performance report via MQTT.

        Note: This data is sent every performance_report_interval (default: 1 second),
        so keep the payload small (recommended: < 1KB) to avoid excessive bandwidth usage.

        Returns:
            dict: Custom information to be included in the MQTT performance report.
                  Empty dict by default.
        """
        return {}



# distutils: language=c++

from libc.stdint cimport int64_t
from libcpp.set cimport set

cdef extern from "../cpp/OrderBookEntry.h":
    cdef cppclass OrderBookEntry:
        OrderBookEntry()
        OrderBookEntry(double price, double amount, int64_t updateId)
        OrderBookEntry(const OrderBookEntry &other)
        OrderBookEntry &operator=(const OrderBookEntry &other)
        double getPrice() const
        double getAmount() const
        int64_t getUpdateId() const

    void truncateOverlapEntries(set[OrderBookEntry] &bid_book, set[OrderBookEntry] &ask_book, const bint &dex)
